#!/usr/bin/perl -w
#Version 4.1.3 modified for ESTminer package
#Program to find alleles in Non-Williams ESTs.  Uses output from
#SNPfind.pl
#Input:	/path/to/snp_alignment file (snp_align_out.txt) 
#		/path/to/Williams haplotype file (snp_est_seqs.txt_haplotypes)
#
#Output:	./NW_alleles (all Non-Williams alleles)
#			./NW_No_LDP_Overlap (NW Ests that did not overlap any Will LDPs)
#			./NW_No_Haps (NW ESTs that did not have a Williams haplotype to compare to)
#			./NW_No_Match (NW ESTs that did not match any LDP - potential new allele)
#
# NOTE: this version recognises contigs with one haplotype and prints out any hit with
#			differences noted
#By:		Rex Nelson
#Completed:	12-06-03
#
#Modified:	12-2-04
#By:		Rex Nelson
#
# Change Log
# Added POD Documentation 12-2-04
# Changed the Output directory from ../OutputFiles/AlleleFinder to ../OutputFiles/AlleleFind


######################################################################################
#
#	PUBLIC DOMAIN NOTICE
#
# This software is "United States Government Work" under the terms of the
# United States Copyright Act.  It was written as part of the authors' offical
# duties for the United States Government and thus cannot be copyrighted.  This
# software is freely available to the public for use without a copyright notice.  
# restrictions cannot be placed on its present or future use.
#
# Although all reasonable efforts have been taken to ensure the accuracy and
# reliability of the software, the United States Department of Agriculture Agricultural 
# Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
# the performance or results that may be obtained by using this software.  
# The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
# merchantability or fitness for any particular purpose.
#
# In any work or product derived from this material, proper attribution of the
# authors as the source of the software should be made.
#######################################################################################

=head1 PUBLIC DOMAIN NOTICE

This software is "United States Government Work" under the terms of the
United States Copyright Act.  It was written as part of the authors' offical
duties for the United States Government and thus cannot be copyrighted.  This
software is freely available to the public for use without a copyright notice.  
restrictions cannot be placed on its present or future use.

Although all reasonable efforts have been taken to ensure the accuracy and
reliability of the software, the United States Department of Agriculture Agricultural 
Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
the performance or results that may be obtained by using this software.  
The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
merchantability or fitness for any particular purpose.

In any work or product derived from this material, proper attribution of the
authors as the source of the software should be made.


=head1 NAME

AlleleFinderV4.1.3.pl

=head1 SYNOPSIS	

	perl AlleleFinderV4.1.3.pl 
	
=head1 DESCRIPTION

AlleleFinderV4.1.3.pl expects to find the input files "../OutputFiles/SNPfind/snp_align_out.txt" and 
"../OutputFiles/HaplotypeSorter/snp_est_seqs.txt_haplotypes" in the indicated subdirectories. The program
produces a number of output files in the "../OutputFiles/AlleleFinder/" subdirectory.

This program compares EST sequences of genotypes other than the inbred strain used in snpfind.  It does this
by comparing all pHaps in a gene family (contig group) to every Other-Genotype EST sequence in the gene-family 
alignment.  If the Other-Genotype EST sequence matches a pHap at all LDP positions in an overlap, it is considered
"associated" with that pHap.  Because of the fragmentary nature of the EST sequences, it is possible for a
particular Other-Genotype EST sequence to be associated with more than one pHap.  Other-Genotype EST sequences which
were associated with only a single pHap sequence are considered to be "unambiguously" associated with a pHap and
their accession numbers are written to the "OG_Unambiguous" file.  Note that the opposite is not true, the pHap which
is unambiguously assoicated with one Other-Genotype EST sequence may also be associated with more than one Other-Genotype 
EST sequence.  Those Other-Genotype sequences which are associated with one or more pHap sequences are written to the
file "OG_alleles".
 


=cut

use strict;


my ($infile,$infile2,$record,$name,$seq,$accn,$longname);
my ($hapnum,$start,$stop,$contig,$hapID,$line,$Qname,$len);
my ($pos,$queryseq,%end5,%end3,$NWcontigs);
my ($query,$hashseq,$longname2,$contig2,$seq2,$start2,$stop2,$accn2);
my ($key,$char,$skipnohap,$nohaps);
my (%haps,@keys,@lines,%seqhash);
my (@lengths,@positions,%prime5,%prime3,%holder,@sortedkeys,$year,$month,$day);
my (@sortedlist,$estname,$estname2,$leader);
my ($outfile,$outfile2,$outfile3,$outfile4,$outfile5);

#get name of input file1 (snp_align_out.txt file)
$infile = "../OutputFiles/SNPfind/snp_align_out.txt";
#get name of input file2 (Inbred genotype haplotype file)
$infile2 = "../OutputFiles/HaplotypeSorter/snp_est_seqs.txt_haplotypes";
$outfile = "../OutputFiles/AlleleFind/OG_alleles";#Other-Genotype ESTs that matched an in-bred haplotype
$outfile2 = "../OutputFiles/AlleleFind/OG_No_LDP_Overlap";#Other-Genotype ESTs that did not overlap any LDPs
$outfile3 = "../OutputFiles/AlleleFind/OG_No_Haps";#Other-Genotype ESTs that did not have a in-bred haplotype to compare
$outfile4 = "../OutputFiles/AlleleFind/OG_No_Match";#Other-Genotype ESTs that did not match the LDPs
$outfile5 = "../OutputFiles/AlleleFind/OG_Unambiguous";#Other-Genotype ESTs that matched only 1 haplotype

open INFILE, "<$infile" or die "Can not open alignment File:$infile because $!\n";
open INFILE2, "<$infile2" or die "Can not open haplotype File:$infile2 because $!\n";

open OUT, ">$outfile" or die "Can not open $outfile because $!\n"; #Other-Genotype ESTs that matched a W haplotype
open OUT3, ">$outfile2" or die "Can not open $outfile2 because $!\n";#Other-Genotype ESTs that did not overlap any LDPs
open OUT4, ">$outfile3" or die "Can not open $outfile3 because $!\n";#Other-Genotype ESTs that did not have a W haplotype to compare
open OUT5, ">$outfile4" or die "Can not open $outfile4 because $!\n"; #Other-Genotype ESTs that did not match the LDPs
open OUT6, ">$outfile5" or die "Can not open $outfile5 because $!\n"; #Other-Genotype ESTs that matched only 1 haplotype

############################################################################################
# Load LDP file (Haplotype file)
############################################################################################
print "Loading LDP file:$infile \n";

#Reset the record separator
$/="\n";

$line = <INFILE2>;
chomp $line;

#split each line into its parts
($longname,$contig,$seq,$start,$stop,$accn) = split /\t/,$line;


#Get the haplotype number out of $longname
if ($longname =~/pHap(\d+)/){
	$hapnum = $1;
	} 
else {
	print "##### 1 Missed hapnum of $longname\n";
	}#end if $longname

$leader = "";	
#Make a space leader the length of $start - 1
#and add it to the beginning of each sequence
for (my $i = 1; $i< $start;$i++){
	$leader .= " ";
	}
my $newseq = "$leader$seq";

#Make the haplotype:sequence combination
$hapID = "$hapnum:$newseq";
#print "$hapID\n";


#Add the start position of the haplotype to %prime5
#using the hapnum as key
$prime5{$hapnum}=$start;

#Add the haplotype:sequcne combo to %holder using
#the hapnum as key
$holder{$hapnum}=$hapID;

#Go through the rest of the file line by line
while ($line = <INFILE2>){


		
	chomp $line;
	
	#split each line into its parts
	($longname2,$contig2,$seq2,$start2,$stop2,$accn2) = split /\t/,$line;

#Skip lines without haplotypes and lines for pHap 900 and 999.
#pHap 900 and 999 are codes for sequences which were coalesced into other sequences
#so we want to ignore them.  pHaps 900 and 999 are used for bookkeeping purposes.
#allows you to follow the fate of each sequence and to identify the ESTs that make up
#any joined sequence.
	if ($longname2 =~ /No/ || $longname2 =~/900/ || $longname2 =~ /999/){
		next;
		}


	#print "Name:$longname2:$contig2:$start2\n";

	#Test to see if the Contig number has changed
	if ($contig eq $contig2){
		
		#Get the haplotype number out of $longname
		if ($longname2 =~/pHap(\d+)/){
			$hapnum = $1;
			} else
				{
				print "##### 2 Missed hapnum of $longname\n";
				
				}#end if $longname
				
		if ($hapnum == 900 || $hapnum == 999){
			next;
			}
	
	$leader = "";	
	#Make a space leader the length of $start - 1
	#and add it to the beginning of each sequence
	for (my $i = 1; $i< $start2;$i++){
		$leader .= " ";
		}
		$newseq = "$leader$seq2";

	
		$hapID = "$hapnum:$newseq";
		#print "$contig:$hapID\n";
		$prime5{$hapnum}=$start2;
		$holder{$hapnum}=$hapID;
		}
	else { #The contig number has changed so we process the contig hashes
		
		#We sort the haplotype numbers by their starting position from
		# left to right
		@keys = keys(%prime5);
		#print "prime5=@keys\n";
		@sortedkeys = sort {$prime5{$a} <=> $prime5{$b}} (keys %prime5);
		#print "sorted keys are:@sortedkeys\n";
		####################################################
		#Load each haplotype:seq combination into %haps    #
		#under the contig name in comma delimited fashion  #
		####################################################

		#start the process by shifting the first name from @sortedkeys
		#and concatinating it to $haps{$contig}
		$key = shift(@sortedkeys);
		#print "Contig:$contig\t$key\t$holder{$key}\n";
		$haps{$contig} = $holder{$key};

		#Go through the remaining keys concatinating each value to 
		#$haps{$contig}
		foreach $key(@sortedkeys){
			#print "Sorted key is:$key\n";
			$haps{$contig} .= ",$holder{$key}";
			}#end foreach
			
		$contig = $contig2;
		$contig2 = "";
		%prime5 = ();
		%holder = ();

		#Get the haplotype number out of $longname
		if ($longname2 =~/pHap(\d+)/){
			$hapnum = $1;
			} 
		else
			{
			print "##### 3 Missed hapnum of $longname2\n";
	
			}#end if $longname
			
		if ($hapnum == 900 || $hapnum == 999){
			next;
			}


		$prime5{$hapnum}=$start2;
		$leader = "";
		for (my $i = 1; $i< $start2;$i++){
			$leader .= " ";
			}
		$newseq = "$leader$seq2";

		$hapID = "$hapnum:$newseq";
		$holder{$hapnum}=$hapID;
		$longname = $longname2;
		
		}#end if $contig eq $contig2

}#end while $line

#if the last line of the haplotype file is a new contig then we need
#to add the values of the contig to %haps - we know this has happened
#if $contig2 == "" when the program gets to this point

if ($contig2 eq "") {
	if ($longname2 =~/pHap(\d+)/){
		$hapnum = $1;
		} else
			{
			print "##### Missed hapnum of $longname\n";
			}#end if $longname
	$leader = "";
	for (my $i = 1; $i< $start2;$i++){
		$leader .= " ";
		}
		$newseq = "$leader$seq";
			
	$hapID = "$hapnum:$newseq";
	$haps{$contig} = $hapID;
	
}#end if $contig2 eq ""

print "LDP file loaded\n";
close INFILE2;

#############################################################################################
# Read in the alignments of each contig group from snp_align_out.txt
##############################################################################################

print "Reading alignment file $infile\n";

#prime the pump
$line = <INFILE>;
chomp $line;
	
($contig,$accn,$seq) = split /\t/, $line;
unless ($seq ne " ") {
	$line = <INFILE>;
	chomp $line;
	($contig,$accn,$seq) = split /\t/, $line;
	}
	
print "###### Processing $contig ######\n";

#lower case any Other-Genotype or Query sequence
#and save in %seqhash
if ($accn =~ /XX/ || $accn =~ /QUERY/){
	$seq =~ tr/[AGCTN]/[agctn]/;
	$seqhash{$accn} = $seq;
	}
if ($accn =~/QUERY/){
	$query = $accn;
	}

while ($line = <INFILE>){

	chomp $line;
	
	($contig2,$accn2,$seq2) = split /\t/, $line;
	
	if (!$accn2){
		$skipnohap++;
		print "No accession detected in $contig2 - skipping\n";
		next;
		}

	if ($contig2 eq $contig){
		if ($accn2 =~ /XX/ || $accn2 =~ /QUERY/){
			$seq2 =~ tr/[AGCTN]/[agctn]/;
			$seqhash{$accn2} = $seq2;
			if ($accn2 =~/QUERY/){
				$query = $accn2;
				}#end if $accn2 =~/QUERY/
			}#end if $accn2 =~ /XX/
		else {
			}
		$contig2 = "Empty";
		}
	else {	
		print "###### Processing $contig #######\n";
		
		&FindAlleles ($contig);
		$NWcontigs++;

		$contig = $contig2;

		$accn = $accn2;
		$seq = $seq2;
		#wipe out old sequences
		%seqhash = ();
		#Start loading the new contig's sequences
		if ($accn =~ /XX/ || $accn =~ /QUERY/){
			$seq =~ tr/[AGCTN]/[agctn]/;
			$seqhash{$accn} = $seq;
			if ($accn2 =~/QUERY/){
				$query = $accn2;
				}#end if $accn2 =~/QUERY/
			}#end if $accn =~ /XX/

		}#end if $contig2 eq $contig
		
}#end while

#process the last contig group - we know it it is in %seqhash
#because $contig2 == Empty

if ($contig2 eq "Empty"){
		print "###### Processing $contig #######\n";
		&FindAlleles ($query);
		$NWcontigs++;
		}

close INFILE;

print "Contigs skipped because of no haplotypes = $skipnohap\n";
print "Number of contigs with Other-Genotype sequences = $NWcontigs\n";
($year,$day,$month) = &GetDate;	
print "AlleleFinderV4.1.2.pl completed $month-$day-$year\n";


##############################################################################
# Begin Subroutines
##############################################################################

#Sub FindAlleles - checks LDPs from each haplotype and logs any SNPs
#  if all LDPs in the overlap are found then any SNPs are capitalized and the sequence
#  added to the appropriate output files

sub FindAlleles (){
	my $Cname = shift;
	print "Cname=$Cname\n";
	my $Qseq = "";
	my $QueryLen = 0;
	my $haplength = 0;
	my %same = ();
	my %snps = ();
	my %LDPmiss = ();
	my $counthap = 0;
	my %ratioHash = ();
	my %HighHapnum = ();
	my %HapHash = ();
	my $snplist = "";
	my $snpct = 0;
	my $flag = 0;
	my %HapOut = ();
	my @HapNums = ();
	my $hapflag = 0;


	unless ($haps{$Cname}) {
		print OUT4 "$Cname\tNo haplotypes\t\n";
		print "#### No haplotypes - Skipping\n";
		$skipnohap++;
		return ; #signifies no haplotypes
		}

	#Get the haplotypes for the contig
	my @haplotypes = split /,/, $haps{$Cname};
	#print "haplotypes are:@haplotypes\n";

	#Get the first haplotype and split it into its parts and
	
	#Get the length of the  alignment by getting the 
	#length of the query sequence
	$Qseq = $seqhash{$query};
	$QueryLen = length($Qseq);
	
	#load up %HapHash
	foreach $line(@haplotypes){
		#Split each line into its parts
		($hapnum,$seq) = split /:/, $line;
		#load each haplotype's sequence into HapHash
		$HapHash{$hapnum} = $seq;
		}
	
	@haplotypes = keys(%HapHash); #hash of num => haplotype seq
	
	my @sorted = sort {$a <=> $b} @haplotypes; #sort haplotype numbers 1 to n
	#print "Haplotype nums=@sorted\n";
	$haplength = length($seq);

	
	
	@keys = keys %seqhash;
	
	if (@keys == 1){
		print "No Other-Genotype Sequences\n";
		return;
		}
	elsif (@keys > 2){ #if more than 1 haplotype
		$hapflag = 1;
		}
	
	print "------ ADPs:\n";
	
	#Go through each Other-Genotype name in seqhash
	foreach $name(@keys){
		if ($name =~ /QUERY/){
			next;
			}
		print "Processing Est $name\n";
		$counthap = 0;
		my $NWseq = $seqhash{$name};
		$NWseq =~ tr/[AGCTN]/[agctn]/;
		my $seqlen = length($NWseq);
		@HapNums = ();
		%HapOut =();
							
		#Go through each haplotype comparing to the
		#Other-Genotype sequence
		
		foreach $hapnum(@sorted){
		
			#print "--- Haplotype $hapnum\n";
			
			$seq = $HapHash{$hapnum}; #the hash of all haplotypes for the group
			
			my $LDPcount = ($seq =~ tr/AGCTN-/AGCTN-/); #count the UC letters (LDPs)
			$counthap++;
			
		
			$haplength = length($seq); #sequence and sequence length of the 
										#haplotype under consideration
			my $possible = 0;
			my $LDPmissed = 0;
			my $LDPmatch = 0;
			my $ratio = .0001;
			$snpct = 0;
			$snplist = "";
			#create a copy of $NWseq to change with UC letters
			my $NWseq2 = $NWseq;
			
			for (my $i = 0;$i< $QueryLen;$i++){
				my $char = "";
				my $nwchar = "";
				
				if ($seqlen < $i+1 || $haplength < $i+1){
					last;
					}
					
				$char = substr($seq,$i,1); #the haplotype character
				
				if ($nwchar = substr($NWseq2,$i,1)){ #the Other-Genotype EST character
					#print "Char=$char\tNwchar=$nwchar\n";
					if ($char eq " " || $nwchar eq " "){ # If spaces 
						next; #get the next $i
						}

					#print "$char\t$nwchar\n";
					if (lc $char ne $nwchar && $char =~ /[AGCTN-]/){ #check for match
						$LDPmissed++; #if not eq increment counter
						next;
						}
						
					if (lc $char ne $nwchar && $char =~ /[agctn-]/){
						$pos = $i + 1;
						$nwchar = uc $nwchar;
						$snpct++;
						
						if ($snplist ne ""){
							$snplist .= ",$pos:$nwchar";
							substr($NWseq2,$i,1) = $nwchar;
							}
						else {
							$snplist .= "$pos:$nwchar";
							substr($NWseq2,$i,1) = $nwchar;
							}
						}#end if lc $char ne $nwchar
						
					if (lc $char eq $nwchar && $char =~ /[AGCTN]/){
						$LDPmatch++;
						
						}
				}#end if (my $nwchar
			}#end for $i
			
			$possible = $LDPmatch + $LDPmissed;
			
			if ($LDPcount != 0 && $LDPmatch > 0){
				#$ratio = $LDPmatch/$LDPcount;
				$ratio = $LDPmatch/$possible;
				}
				
			#print "$name\t$LDPmatch\t$LDPmissed\t$LDPcount\n";
			
			if ($LDPmatch == 0 && $LDPmissed > 0){
				$LDPmiss{$name}++;
				}
				
			if ($ratio == 1 || $possible == 0){

				if ($possible == 0){
					$flag++;
					if ($hapflag == 0){ #there was only one haplotye
						$HapOut{$hapnum} = "$seq\t$LDPmatch\\$possible\t$LDPcount\t$snpct\t$snplist";
						print "  $hapnum\t$LDPmatch\\$possible\t$LDPcount\t$snpct\n";
						#We now want to keep any Other-Genotype differences to a contig with one pHap so
						#we assign the value of NWseq2 to NWseq
						$NWseq = $NWseq2;
						}
					}
				else {
					$HapOut{$hapnum} = "$seq\t$LDPmatch\\$possible\t$LDPcount\t$snpct\t$snplist";
					print "  $hapnum\t$LDPmatch\\$possible\t$LDPcount\t$snpct\n";
					#The changes made to NWseq2 are good so want them incorporated into
					#NWseq so we assign the value of NWseq2 to NWseq
					$NWseq = $NWseq2;
					}
				
				}#end if $ratio == 1
			
		}#end foreach $hapnum

		if ($counthap == $flag){
			print "$name did not overlap any LDPs\n";
			print OUT3 "$name\t$Cname\t$NWseq\n";
			$flag = 0;
			}
		else {
			$flag = 0;
			}
			#print "Haplotype count=$counthap  Missed LDPs=$LDPmiss{$name}\n";
			#if LDPs were missed and the haplotpye count = number of haplotypes with LDP miss
		if ($LDPmiss{$name} && $counthap == $LDPmiss{$name}){
			print "$name missed all LDP's\n";
			print OUT5 "$name\t$Cname\t$NWseq\n";
			}
			
		if (%HapOut){
			
			print OUT "$name\t$Cname\t000\t$NWseq\n";
			@HapNums = keys (%HapOut);
			if (@HapNums == 1){
				$hapnum = $HapNums[0];
				print OUT6 "$name\t$Cname\t$hapnum\t$HapOut{$hapnum}\n";
				}
			my @sorted2 = sort {$a <=> $b} @HapNums;
			foreach $hapnum(@sorted2){
				$seq = $HapOut{$hapnum};
				print OUT "        \t$Cname\t$hapnum\t$seq\n";	
				}#end foreach
			}#end if %HapOut
		
			
	}#end foreach $name
	
}#end sub FindAlleles

###############################################################
# Sub Fasta2Seq - decomposes a fasta record into defline and sequence
###############################################################
sub Fasta2Seq {

  my $record = "";
  my $DefLine = "";
  my $seq = "";
  my $line = "";
  my @lines = {};

  $record = shift(@_);

  @lines = split(/\n/, $record);

  $DefLine = shift(@lines);

  $seq = join("", @lines);

  return $DefLine,$seq;

}#end sub Fasta2Seq

################################################################
# Subroutine GetDate - Composes a date stamp
################################################################

sub GetDate {
my $year = 0;
my $month = 0;
my $day = 0;
#get date information for a day stamp
($day,$month,$year)=(localtime)[3,4,5];
return $year+1900, $month+1, $day;

}#end sub GetDate

