#!/usr/bin/perl -w
#Program to identify haplotypes
#based on LDP pattern in snp_est_seqs.txt
#output file of snpfindV14.0.2
#This file AKA David4.7.pl
#was modified for ESTminer
#version David4.7.1
#By: Rex Nelson
#Last Modified: 12-2-04
#By: Rex Nelson
#
#CHANGE LOG
# Added POD Documentation 12-2-04


######################################################################################
#
#	PUBLIC DOMAIN NOTICE
#
# This software is "United States Government Work" under the terms of the
# United States Copyright Act.  It was written as part of the authors' offical
# duties for the United States Government and thus cannot be copyrighted.  This
# software is freely available to the public for use without a copyright notice.  
# restrictions cannot be placed on its present or future use.
#
# Although all reasonable efforts have been taken to ensure the accuracy and
# reliability of the software, the United States Department of Agriculture Agricultural 
# Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
# the performance or results that may be obtained by using this software.  
# The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
# merchantability or fitness for any particular purpose.
#
# In any work or product derived from this material, proper attribution of the
# authors as the source of the software should be made.
#######################################################################################

=head1 PUBLIC DOMAIN NOTICE


 This software is "United States Government Work" under the terms of the
 United States Copyright Act.  It was written as part of the authors' offical
 duties for the United States Government and thus cannot be copyrighted.  This
 software is freely available to the public for use without a copyright notice.  
 restrictions cannot be placed on its present or future use.

 Although all reasonable efforts have been taken to ensure the accuracy and
 reliability of the software, the United States Department of Agriculture Agricultural 
 Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
 the performance or results that may be obtained by using this software.  
 The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
 merchantability or fitness for any particular purpose.

 In any work or product derived from this material, proper attribution of the
 authors as the source of the software should be made.

=head1 NAME

	HaplotypeSorterV1.1.pl
	
=head1 SYNOPSIS
	
	perl HaplotypeSorterV1.1.pl

=head1 DESCRIPTION

	This program reduces the redundancy in the output from snpfindV14.0.7. It takes as an input
	the snpfindV14.0.7.pl output file snp_est_seqs.txt which it expects to find in the subdirectory
	"../OutputFiles/SNPfind/snp_est_seqs.txt". Output files from this program are written to the 
	"../OutputFiles/HaplotypeSorter" subdirectory. Note that a considerable amount of commentary is
	also written to the terminal. You may want to capture the commentary for quality control purposes.

=head1 ALGORITHM GENERAL DESCRIPTION
	
	The program processes the sequences from each gene family (contig group). The program compares 
	each sequence based on its LDP pattern. If the LDP
	pattern of two sequences match, the longer of the sequences is used as a representative of that 
	LDP pattern. If one of the sequences' LDP pattern was a complete subset of the LDP pattern of 
	another sequence, then it is considered to be contained within the larger sequence and the larger
	sequence will represent the smaller sequence. In gene families with a single sequence, any overlapping
	sequences are assembled together. In all other cases, overlapping sequences are NOT assembled together.
	(See the ESTminer publication for an explaination) The resulting sequences are then termed Potential
	Haplotypes (pHaps) representing the "haplotype" of a gene. Since we used a highly inbred line in our
	LDP identification, all loci are considered HOMOZYGOUS and thus each LDP pattern represents the sequence
	of gene in a family of highly related sequences (gene family). 

=cut

use strict;

my ($filename,$outfile,$outfile2,$line,$contig,$est,$name,$seq,$HaplotypeCt,$contig2,$longest);
my ($pos5,$pos3,$UCct,$Qlen,$num,$hapct,$longname,$outfile3,$pseudoname,$ct,$name1,$name2);
my (@ests,@list,@sortednames,@same_ct,@finalhaplotypes1,@outnames,@finalhaplotypes2);
my (%hash,%prime5,%prime3,%first,%elementct,%UChash,%last,%pseudonames,@haplolist,%copy,@same);
my ($outfile4,$UClist,$result,%UClist,%endslist,@endlist5,@endlist3,%same,@holder2);
my (@holder3,$outfile5);

$filename = "../OutputFiles/SNPfind/snp_est_seqs.txt";

open INFILE, "<$filename" or die "Can not open input file $filename because $!\n";

$outfile = "../OutputFiles/HaplotypeSorter/snp_est_seqs.txt_haplotypes";
$outfile2 = "../OutputFiles/HaplotypeSorter/coalesced_ests.txt";
$outfile3 = "../OutputFiles/HaplotypeSorter/consensus_ests.txt";
#$outfile4 = "../OutputFiles/HaplotypeSorter/pseudonames.txt";
$outfile5 = "../OutputFiles/HaplotypeSorter/uninformative.txt";

#open output file for writing, clears out any old copies and makes sure it can be opened
open OUT1, ">$outfile" or die "Can not open output file $outfile because $!\n";
open OUT2, ">$outfile2" or die "Can not open output file $outfile2 because $!\n";
open OUT3, ">$outfile3" or die "Can not open outputfile $outfile3 because $!\n";
#open OUT4, ">$outfile4" or die "Can not open outputfile $outfile4 because $!\n";
open OUT5, ">$outfile5" or die "Can not open outputfile $outfile5 because $!\n";

=head2 Read in one contig group at a time
	
	Read in each line from the input file and check for a change in the contig number.
	If the contig number changes process the gene family sequences
	
=cut

while ($line = <INFILE>){
	chomp($line);
	@ests = ();
	%first = ();
	%hash = ();
	%elementct = ();
	%UClist = ();
	%last = ();
	$ct =0;
	$hapct = 1;
	@finalhaplotypes2 = ();
	%endslist = ();
	@endlist5 = ();
	@endlist3 = ();
	@holder2 = ();
	@holder3 = ();

	#split the line on tabs into its constituent parts
	#assign each part to @ests
	@ests = split(/\t/, $line);
	
	#shift the first part of the line (contig) into $contig
	#leaving the sequence information for every est in the array
	$contig = shift(@ests);
	#pop the query length off array
	$Qlen = pop(@ests);
	print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n";
	print "### Processing $contig ###\n";
	print "@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@@\n";
	if ($ests[0] eq " "){
		print "No ESTs Found - Skipping\n";
		$longname = &PadContig($contig);
		print OUT1 "No ESTs\t$contig\t \t \t \t \t \n";
		next;
		}
	

	#for each est in the array 
	foreach $est(@ests){
		#split the est information on commas 
		#producinge the EST name, EST sequence, sequence start position, sequence stop position
		($name,$seq,$pos5,$pos3) = split(/,/, $est);
		$ct++;
		#Make a pseudoname list for each EST
		$pseudoname = "grp$contig" . "_$ct";
		$pseudonames{$pseudoname}=$name;
		#Add EST name and sequence to %hash
		#$hash{$pseudoname} = $seq;
		$hash{$name} = $seq;
		#Add start position to %prime5
		#$prime5{$pseudoname}=$pos5;
		$prime5{$name}=$pos5;
		#Add stop position to %prime3
		#$prime3{$pseudoname}=$pos3;
		$prime3{$name}=$pos3;
		#print "Contig=$contig\tName=$name\t$pseudoname\t5prime=$pos5\t3prime=$pos3\n";
		print "Contig=$contig\tName=$name\t$name\t5prime=$pos5\t3prime=$pos3\n";
		#print OUT4 "$contig\t$pseudoname\t$name\n";
		}#end foreach $est
		
	$HaplotypeCt = 1;

	@list = keys (%hash);
	&Look4UC(@list);
	@list = keys(%hash);
###########################################################
	my $linect = 1;
	my $linect2 = 0;
	my @items = ();
	my $firstitem = "";
	my $first = 0;
	my $first2 = 0;
	my $char = "";
	my $char2 = "";
	@holder2 = ();
	my $Flag = 0;
	my @temp = ();
	my @holder = ();
	my $seq1 = "";
	my $seq2 = "";

	#haplolist is a list of all names sorted by size and LDP start
	@haplolist = &SortByLDP(@list);
	print "Haplolist returned=@haplolist\n";

	@list = @haplolist;

	#while @list exists
	while (@list){
		#get the line count of @list
		#print "List=@list\n";	
	
		#clear out @same
		@same = ();
	
		#get the first name in the list (smallest 5prime LDP)
		$name = shift@list;
		print "Name=$name\n";
	
		#put name into @same
		push @same, $name;

		#Get the LDP pattern from %UChash
		$seq1 = $UChash{$name};

		#find out if we need to treat the name as a sequence or LDP List
		#If there are no LDPs in a group, %UChash will be empty. $seq1 will
		#be assigned undef and thus this test will fail indicating that
		#there are no capital letters and all we have in the group is sequence
		#if ($seq1 =~/[agctnAGCTN-]/){
		if ($seq1 =~/[agctnAGCTN-]/){
			#Seq is a list
			
			print "---- LDP LIST ----\n";
			my $flag5 = 0;
			#go through remaining names one by one
			foreach $name2(@list){

				#initalize $result for each name
				$result = "0";
			

				#get the LDP sequence of name2
				$seq2 = $UChash{$name2};
		
				#test for an exact match of LDP Pattern
				#if matches add name to @same else add to
				#@holder
			
				print "  Sending $name and $name2 to Sub Contained\n";
				#use &Contained for LDP list to test for an exact match
				$result = &Contained($name,$name2);
				
				print "----------- Result of Contained =$result\n";
				if ($result == 1){
		
					push @same, $name2;
					}
				elsif ($result == 2 ){ #name was a subset of name2
					#if ($flag5 == 0){
						#shift @same; #remove name from list (its the first item)
						#$flag5 = 1;
						#push @same, $name2
						#}
					#print OUT2 "$contig\t$name\t$name2\n";
					push @holder, $name2;
					}
				else { #name2 was different
					push @holder, $name2;
					}
				}#end foreach $name2
			if (@same){
				print "  Sending @same\nto OverLap2\n";	
				$result = &OverLap2(@same);
				print "  Result of OverLap2 =$result!\n";
	
				#put the result (allways a sequence) in @holder2
				#@holder2 persists through the iterations
				push @holder2, $result;
				}
			
			#assign @list the values in @holder for next iteration
			@list = @holder;
		
			#delete contents of @holder for next iteration
			@holder = ();
			
			}
		else {
			#Seq is a sequence
			
			print "---- Sequence -----\n";
			#the sequences are sequences so we just
			#overlap all of the sequences using OverLap
			push @list, $name; #we took $name out of list so need to put it back
			print "\n\n*********** NO UC -- Sending all sequences to OverLap\n\n";
			@holder2 = &OverLap(@list);
			print "OverLap returned:\n@holder2\n";
			@list = ();
			}
	}#end While @list;
	
	@holder = ();
	@list = @holder2;
	while (@list){
		$name = shift(@list);
		$result = 0;
		if (@list){
			foreach $name2(@list){
				$result = 0;
				$result = &Contained($name,$name2);
				print "last result =$result\n";

				if ($result == 2){
					print OUT2 "$contig\t$name\t$name2\n";
					print OUT5 "$contig\t$name\tContained within\n";
					my $longname = &PadContig($contig);
					print OUT1 "$longname-pHap900\t$contig\t$hash{$name}\t$prime5{$name}\t$prime3{$name}\t$name\n";
					last;
					}#end if $result == 2
				
				}#end foreach $name2
				
			}#end if @list
		unless ($result == 2){
			push @holder, $name;
			}
		}#end while @list
	&PrintOut(@holder);		

	
}#end while $line = <INFILE>

close OUT1;
close OUT2;
close OUT3;
#close OUT4;
close OUT5;
close INFILE;

my ($year,$month,$day) = &GetDate;

print "##### Program HaplotypeSorterV1.1.pl finished file $filename $month-$day-$year######\n";	
exit;

##################################################################################
#	Begin Subroutines															 #
##################################################################################
=head2 Check Each Sequence in a Family for Upper Case Letters
	
	We check each sequence in a family group for capital letters. If no capital
	letters are encountered in a sequence, we remove the name of the sequence from
	the hash (%hash) because it is uninformative in the sense that we can not assign it to
	a particular pHap sequence. If there are no upper case letters in any of the 
	sequences we remove all the sequences from the hash (%hash)
	
=cut

######################################################################
#  Look for upper case letters in a list of sequences
######################################################################	
sub Look4UC(){
	my @LookList = (@_);
	my ($name,$ct,$UCpos,$char,$first,$start,$stop,$last);
	my ($seqlen);
	my @holder = ();
	my $UCflag = 0;
	print "#########################  Sub Look4UC  ###########################\n";
	#print "Looking for UC in @LookList\n";
	foreach $name(@LookList){ #looking to see if there are any CAPs at all
		$seq = $hash{$name};
		if ($seq =~ /[AGCTN]/){
			$UCflag = 1; #if a cap is encountered set UCflag and quit
						 #if no cap is encountered then we will have 1 pHap an dont count gaps
						 #as an LDP
			last;
			}
		}#end foreach
	foreach $name(@LookList){
		$seq = $hash{$name};
		my $len = length($seq);
		#print "Name is $name\tSeq is $seq\n";
		$start = $prime5{$name};
		$stop = $prime3{$name};
		$first = 0;
		$last = 0;
		my $uclist = "";
		$ct = 0;
		$seqlen = length($seq);
		
		if ($UCflag == 1){
			if ($seq =~ /[AGCTN-]/){		
				for (my $i = 0; $i<$seqlen;$i++){
					$char = substr($seq,$i,1);
					if ($char =~/[AGCTN-]/){
						$UCpos = $i + $start;
						#print "UC is $char\n";
						if ($uclist eq ""){
							$first{$name} = $UCpos;
							$uclist .= "$UCpos:$char";
							$ct++;
							}
						else {
							$uclist .= ",$UCpos:$char";
							$ct++;
							}#end if $uclist
						}#end if $char
					}#end for
			
				$last{$name} = $UCpos;
				}
			else {
				print "$name had no LDPs\n";
				$uclist = "0:X";
				$first{$name} = $start;
				$last{$name} = $stop;
				$ct = $len;
				#add name to @holder
				push @holder, $name;
				}#end if $seq =~ /[AGCTN-]/
			}
		else {
			print "$name had no LDPs\n";
			$uclist = "0:X";
			$first{$name} = $start;
			$last{$name} = $stop;
			$ct = $len;
			#add name to @holder
			push @holder, $name;
		
			}
		
		$elementct{$name} = $ct;
		$UChash{$name}=$uclist;
		
		#print "Name=$name\tUClist=$uclist\t$ct\n";
		#print "First=$first{$name}\tLast=$last{$name}\n";
		}#end foreach $name
		
		#check to see if there are any sequences with UC's by
		#comparing original list with holder list -- if equal
		#there are no UC's in the sequence so we don't do anything
		#if there are some sequences with UC we remove only the ones
		#in the holder list
		if (@LookList != @holder){
			foreach $name(@holder){
				print "$name removed - no UC\n";
				print OUT5 "$contig\t$name\tNo LDPs\n";
				my $longname = &PadContig($contig);
				print OUT1 "$longname-pHap999\t$contig\t$hash{$name}\t$prime5{$name}\t$prime3{$name}\t$name\n";
				delete $hash{$name};
				}
			}
		print "Exiting Look4UC\n";
		return;
	}#end sub Look4UC

=head2 Subroutine SortByLDP

	Orders the sequences first by the number of LDPs it contains and then by 5prime
	start of the sequence such that the shorter sequences in terms of LDPs and 5prime
	start site are processed first. Return the sorted list
	
=cut

################################################################################
# sub SortByLDP
################################################################################
sub SortByLDP(){
	my @inlist = (@_);
	my @finalgrplist = ();
	my $name1 = "";
	my $name2 = "";
	print "SortByLDP inlist=@inlist\n";
	
	if (@inlist == 1){
		push @finalgrplist, @inlist;
		print "################## SortByLDP got one sequence #################\n";
		return @finalgrplist;
		}

	print "######################### Sorting by LDP  ##########################\n";
	#sort all of the names %element by value (number of elements in each list)
	@sortednames = sort {$elementct{$a} <=> $elementct{$b}} @inlist;
	#print "Sorted names = @sortednames\n";
	my @grplist = ();
	#collect all of the names with the same number of elements
	$name1 = shift(@sortednames);
	push @grplist, $name1;
	#print "Grplist=@grplist\n";
	foreach $name2(@sortednames){
		#if ($name2 eq ""){
		#	next;
		#	}
		if ($elementct{$name1} == $elementct{$name2}){
			#print "Element count of $name1 = count of $name2\n";
			push @grplist, $name2;
			#print "Grplist=@grplist\n";
			$name2 = "empty";
			}
		else {
			#sort @grplist by starting position
			my @sortedgrplist = sort {$first{$a} <=> $first{$b}} @grplist;
			#print "Sortedgrplist =@sortedgrplist\n";
			#Add to @finalgrplist the sorted grplist
			push @finalgrplist, @sortedgrplist;
			#print "finalgrplist plus sortedgrplist = @finalgrplist\n";
			#Wipe out the old list
			@grplist = ();
			#assign the value of name2 to name1
			$name1 = $name2;
			#assign the value of "empty" to name2
			$name2 = "empty";
			#add name1to new grplist
			push @grplist, $name1;
			}
		}#end foreach $name2
		
	if (@grplist){
		push @finalgrplist, @grplist;
		}

	return @finalgrplist;

}#end sub SortByLDP
=head2 Subroutine Contained

	This routine compares two sequences in a group and determines if the sequences
	begin and end at the same place in terms of the LDP pattern. If they begin and end at
	the same place they are then inspected for the presence of "N" characters. If they differ
	only by Ns, we will put them together.  If they have Ns we will send them to subroutine
	CompleteContain. Else we will use the perl function "eq" to compare them. Routine returns
	1 (seq1 matches seq2) or 0 (seq1 does not match seq2)

=cut

##################################################################################
#	Subroutine to see if $seq1 is contained within $seq2		 #																 #
##################################################################################
sub Contained() {
	my $name = shift;
	my $name2 = shift;
	
	#Get the start and stop positions for $name
	my $start = $first{$name};
	my $stop = $last{$name};
	my $seq1 = $UChash{$name};
			
	print "###################### Sub Contained  #######################\n";
	print "Comparing $name to $name2\n";
	#get the sequnce and start/stop positions for $name2
	my $start2 = $first{$name2};
	my $stop2 = $last{$name2};
	my $seq2 = $UChash{$name2};
	#print "Name=$name\tFirst=$start\tLast=$stop\tList=$seq1\nName=$name2\tFirst=$start2\tLast=$stop2 List=$seq2\n";
	
			
	#determine if LDPs of name and name2 start at the same place
	if ($start == $start2 && $stop2 == $stop){ ################################################
		#print "$name and $name2 start at the same place\n";
		#print "$name and $name2 start and stop at the same place\n";
		#if it could, check its sequence for N's
		if ($seq1 =~/[nN]/ || $seq2 =~/[nN]/){
			print "Sequences have N's\nSubmitting to CompleteContain\n";
				
			#if N's are found then we will have to check each
			#character in the overlap to see if they are the
			#same by submitting to sub CompleteContain
			
			$result = &CompleteContain($name,$seq1,$name2,$seq2);

			#if the result is "1" then all characters matched
			if ($result == 1){
				print "$name matches $name2\n";
				return "1";
				}#end if $result == 1
			else {
				return "0";
				}
			}				
		else {
			#else there were no N's in the sequence so we
			#use eq to compare both
			
			print "$name2 does not have Ns \n";
			print "Checking for a match\n";
			if ($seq2 eq $seq1){
				print "$name equals $name2\n";
				return "1";
				}
			else {
				return "0";
				}#end if ($seq2 eq $seq1)
			}#end if $seq2 =~ /[nN]/
		}
	elsif ($seq2 =~ /$seq1/) {
		print "$name2 contains $name\n";
		#print OUT2 "$contig\t$name\t$name2\n";
		return "2";
		}
	else{
		print "$name does not start and stop at the same place as $name2\n";
		return "0";
		}#end if $start>=$start2

	
	}#end &Contained
	

=head2 Subroutine CompleteContain
	
	This routine takes two input sequences (LDP patterns) and determins if they
	are the same. It does this manually by inspecting each LDP character. This 
	is necessary to allow sequences which vary only by unknown characters (N) to
	be coalesced into a single sequence. The routine takes two sequence names and the
	actual sequence as input. The routine returns 1 (seq1 eq seq2) or 0 (seq1 ne seq2)
	
=cut
####################################################################
# Sub CompleteContain - routine to manually search an overlapping
# region for characters that do not match
####################################################################

sub CompleteContain () {

my $name = shift;
my $seq1 = shift;
my $name2 = shift;
my $seq2 = shift;
my $len = length($seq1);
my $Flag = 0;
my %LDPs = ();
my $pos = 0;
my $char = "";
my $item = "";

print "########## Complete Contain ##############";

#We know that seq1 is shorter than seq2 so we start at
#sequence and start the comparison with position 0 and compare characters
#If any characters differthe sequnces are not the same so we will return 0.  
#If they are all the same we will return 1.

print "CompleteContain got:$name\t$name2\n";
#print "CompleteContain got:\n$name\n$seq1\n$name2\n$seq2\n";

my @seq1 = split /,/, $seq1;
my @seq2 = split /,/, $seq2;

#since we know they start at the same place they have to be the same length
#to be considered
my $len1 = @seq1;
my $len2 = @seq2;

#Load up hash LDPs with pos => char info from seq1	
foreach $item(@seq1){
	($pos,$char) = split /:/, $item;
	$LDPs{$pos} = $char;
	}#end foreach $item
	
foreach $item(@seq2){
	($pos, $char) = split /:/, $item;
	if (my $char2 = $LDPs{$pos}){
		if ($char =~ /[nN]/ || $char2 =~ /[nN]/){
			next;
			}
		elsif ($char ne $char2){
			$Flag++;
			}#end if $char eq $char2
		}#end if my $char2
	}#end foreach $item

if ($Flag > 0){
	print "CompleteContain found $Flag mismatches\n";
	return "0";
	}
else {
	return "1";
	}
	
}#end sub CompleteContain

=head2 Subroutine OverLap

	Routine to join overlapping sequences in gene families with a single haplotype
	ie no capital letters. Takes as input an array of sequence names and returns the
	larger overlapped sequence. Keeps track of all names coalesced into the new sequence
	by writing to consensus.txt
	
=cut

################################################################################
# Sub OverLap - make a consensus of lc sequences 
################################################################################
sub OverLap(){


my @names = (@_);
my $consensus = "";
my @joined = ();
my $joinedlist = "";
my $name1 = "";
my $name2 = "";
#my @outnames = ();
my $seq = "";
my @seqs = ();
my @sortednames2 = ();
my $n = 0;


print "###############   Sub OverLap  ###################\n";
print "Sub OverLap names:@names\n";
if (@names < 2){
	print "Sub OverLap got 1 name\n";
	$name2 = shift(@names);
	#print "$name2\t$hash{$name2}\n$prime5{$name2}\t$prime3{$name2}\n";
	return $name2;
	}


#sort the names by starting position with the left most first
@sortednames2 = sort {$prime5{$a} <=> $prime5{$b}} @names;								  
#print "Sorted names2=@sortednames2\n";

#Get index of last item
$n = $#sortednames2;

#get start of 5prime most seq
my $beginname = $sortednames2[0];
my $begin = $prime5{$beginname};
#the last 3prime most end
my $last = 0;
#pad the front of each seq with spaces
#and add seq to @seqs
foreach $name(@sortednames2){
	my $end = $prime3{$name};
	#find the 3prime most end
	if ($end > $last){
		$last = $end;
		}
	}
#print "BEGIN=$begin  LAST=$last\n";
foreach $name(@sortednames2){
	#print out each name to _haplotype file
	my $longname = &PadContig($contig);
	print OUT1 "$longname-pHap900\t$contig\t$hash{$name}\t$prime5{$name}\t$prime3{$name}\t$name\n";

	my $leader = "";
	my $start = $prime5{$name};

	my $offset = ($start-$begin);
	
	for (my $i = 0;$i < $offset;$i++){
		$leader .= ' ';
		}
	#print "OFFSET=$offset  Leader=!!$leader!!\n";
	 my $seq1 = "$leader" . "$hash{$name}";
	# print "$name  $seq1\n";
	push @seqs, $seq1;
	$hash{$name} = $seq1;
	#Add name to joinedlist
	if ($joinedlist eq ""){
		$joinedlist .= "$name";
		}
	else {
		$joinedlist .= ",$name";
		}
	}#end foreach $name

print "Begin=$begin End=$last\nJoined List=$joinedlist\n";
my $flag = 0;
my %bases = ();
#go through each column in the alignment	
#for (my $i=0;$i<=($last - $begin);$i++){
for (my $i=0;$i<= $last;$i++){
	my $index = 0;
	
	#go through each name and get the first non-N, Non-space char
	#At this point all sequences should be exactly the same except for
	#N's
	my $char = "";
	my $char2 = "";
	foreach $name(@sortednames2){
			if ($i > length($seqs[$index])){
				$index++;
				next;
				}
			if ($char = substr($seqs[$index],$i,1)){
				if ($char =~ /[agctAGCT-]/){
					$char2 = $char;
					$flag = 1;
					last;
					}
				elsif ($char =~/[nN]/){
					$index++;
					$char2 = $char;
					$flag = 1;
					}
				elsif ($char eq " "){
					$index++;
					}
				$bases{$char}++;
				}
			else {
				$index++;
				
				}
			}#end foreach
	my @bases = keys %bases;
	if (@bases == 1 && $char eq " " && $flag == 1){
		$char2 = "+";
		}
	$seq .= "$char2";
	}#end for
	
#makes a new name for the consensus
my $compname = "$contig:$hapct";
print "New Name=$compname\n";
print "Begins=$begin  Ends=$last\n";
$hapct++;
#put name and sequence into %hash
$hash{$compname} = $seq;
$prime5{$compname} = $begin;
$prime3{$compname} = $last;
&Look4UC($compname);
#print out joined list to OUT3 (consensus.txt)
print OUT3 "$contig\t$compname\t$joinedlist\n";
print "$contig\t$compname\t$joinedlist\n";
return ($compname);
}#end sub OverLap

=head2 Subroutine OverLap2
	
	This routine takes a list of sequences with overlapping LDP patterns and returns
	the sequence with the longest physical length as a representitive of the group (LDP pattern)

=cut
	
#################################################################################
# sub OverLap2 - Takes LDP Overlaps and returns the longest EST as a representative
#################################################################################
sub OverLap2(){


my @names = (@_);
my $consensus = "";
my @joined = ();
my $joinedlist = "";
my $name1 = "";
my $name2 = "";
#my @outnames = ();
my $seq = "";
my @seqs = ();
my @sortednames2 = ();
my $n = 0;


print "###############   Sub OverLap2  ###################\n";
print "Sub OverLap2 names:@names\n";
if (@names < 2){
	print "Sub OverLap2 got 1 name\n";
	$name2 = shift(@names);
	#print "$name2\t$hash{$name2}\n$prime5{$name2}\t$prime3{$name2}\n";
	return $name2;
	}


#sort the names by starting position with the left most first
@sortednames2 = sort {$prime5{$a} <=> $prime5{$b}} @names;								  
#print "Sorted names2=@sortednames2\n";

#Get index of last item
$n = $#sortednames2;

my $begin = 0;
my $last = 0;
my $longestname = "";
my $biglength = 0;
#pad the front of each seq with spaces
#and add seq to @seqs
foreach $name(@sortednames2){
	#get the start and stop of each name
	$last = $prime3{$name};
	$begin = $prime5{$name};
	my $size = $last - $begin;
	#find the longest sequence
	if ($size > $biglength){
		$longestname = $name;
		$biglength = $size;
		}#end if $size
	}#end foreach $name
print "Biglength=$biglength Longestname=$longestname\n";	
print "BEGIN=$begin  LAST=$last\n";
foreach $name(@sortednames2){
	if ($name ne $longestname){
		#Add name to joinedlist
		if ($joinedlist eq ""){
			$joinedlist .= "$name";
			}
		else {
			$joinedlist .= ",$name";
			}
		my $longname = &PadContig($contig);
		print OUT1 "$longname-pHap900\t$contig\t$hash{$name}\t$prime5{$name}\t$prime3{$name}\t$name\n";
		}#end if $name ne $longestname
	}#end foreach $name

print "Joined List=$joinedlist\n";
	
print "Longest Name=$longestname\t";
print "Begins=$begin  Ends=$last\n";

#print out joined list to OUT3 (consensus.txt)
print OUT3 "$contig\t$longestname\t$joinedlist\n";
print "$contig\t$longestname\t$joinedlist\n";
return ($longestname);
}#end sub OverLap2


################################################################
# Subroutine GetDate - Composes a date stamp
################################################################

sub GetDate {
my $year = 0;
my $month = 0;
my $day = 0;
#get date information for a day stamp
($day,$month,$year)=(localtime)[3,4,5];
return $year+1900, $month+1, $day;

}#end sub GetDate

##################################################################
# Subroutine PadContig - zero pads the contig number
##################################################################
sub PadContig (){
	
	my $contigname = shift;
	$contigname =~/Contig(\d+)/;
	my $contignum = $1;
	my @cdigits = split //, $contignum;
	my $digitct = scalar @cdigits;
	my $zeros = "";
	if ($digitct == 1) { $zeros = "0000";}
		elsif ($digitct == 2) {$zeros = "000";}
		elsif ($digitct == 3) {$zeros = "00";}
		elsif ($digitct == 4) {$zeros = "0";}
	else {
		$zeros = "";
		}
		
	my $designator = "Gm-ctg$zeros$contignum";
			
	return $designator;
}

##################################################################
# Subroutine PadNumber - zero pads the haplotype number
##################################################################
sub PadNumber (){

	my $Hnum = shift;
	my $zeros = "";
	my @hapdigits = split //, $Hnum;
	$hapct = scalar @hapdigits;
	if ($hapct == 1) {$zeros = "00";}
		elsif ($hapct == 2) {$zeros = "0";}
		else {$zeros = "";}
	my $designator = "-pHap$zeros$HaplotypeCt";
			
	return $designator;
}


#################################################################################
# Sub PrintOut - formats sequences in a list and writes to a file
#################################################################################
sub PrintOut (){

	$HaplotypeCt = 0;
	my @Inlist = (@_);

	#go through the final list and format for printing
		foreach my $haplo(@Inlist){
			if ($hash{$haplo}){
				#print OUT1 "##### $contig Haplotype Consensus $HaplotypeCt\n";
				$HaplotypeCt++;
				my $leader = "";#number of spaces to pad the front of the sequence
			
				#add a space to $leader for every position in front of the 5' start
				#for (my $i=0;$i<$prime5{$haplo};$i++){
				#	 $leader .= " ";
				#	}
			
				#format the output string
				#determine the zero padding for the contig number
				
				my $contigdesignator = &PadContig($contig);
			
				#determine the zero padding for the haplotype number
				
				my $hapdesignator = &PadNumber($HaplotypeCt);
			
				my $longname = "$contigdesignator$hapdesignator";
			
				print OUT1 "$longname\t$contig\t$leader$hash{$haplo}\t$prime5{$haplo}\t$prime3{$haplo}\t$haplo\t\n";
			}#end if
		}#end foreach
}#end sub PrintOut

	