#!/usr/bin/perl -w
#
#snpfindV14.0.8.pl
#
#MODIFIED VERSION OF SNPFINDV14.0.7.PL 
#This version was modified for ESTminer.
#This version uses a database of inbred and non-inbred Est sequences.
#The Non-inbred Ests are only considered when a inbred EST needs to
#be validated.  The Non-inbred Ests are then queried for match.  If no 
#match is found(orthere are no non-inbred sequences) then the EST name is added 
#to the snp_rejects.txt file.  This program assumes that the Contig sequences are included
#in the Blast database and ignores them.
#
#Note:  This program does not remove sequences that induce gaps in the Query Sequence.
#		This program DOES NOT REMOVE repetitive sequences, uses the first line only.  
#		Checks for internal spaces and replaces them with "-".
#		Encodes gaps as GapStart:D:GapEnd in snp_out.txt
#		Fixed GapEnd problem 11-4-2004
#		Fixed repetitive sequences being prematurely truncated 11-4-2004
#		Added POD Documentation 11-26-2004
#
#Input is ../InputFiles/BlastOut ../InputFiles/BlastDB
#
#
#
#
#
#Started			11-21-03
#
#By					Rex Nelson
#
#Last Modified:		11-26-04
#
#By:				Rex Nelson
#



######################################################################################
#
#	PUBLIC DOMAIN NOTICE
#
# This software is "United States Government Work" under the terms of the
# United States Copyright Act.  It was written as part of the authors' offical
# duties for the United States Government and thus cannot be copyrighted.  This
# software is freely available to the public for use without a copyright notice.  
# restrictions cannot be placed on its present or future use.
#
# Although all reasonable efforts have been taken to ensure the accuracy and
# reliability of the software, the United States Department of Agriculture Agricultural 
# Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
# the performance or results that may be obtained by using this software.  
# The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
# merchantability or fitness for any particular purpose.
#
# In any work or product derived from this material, proper attribution of the
# authors as the source of the software should be made.
#######################################################################################
=head1 PUBLIC DOMAIN NOTICE

This software is "United States Government Work" under the terms of the
United States Copyright Act.  It was written as part of the authors' offical
duties for the United States Government and thus cannot be copyrighted.  This
software is freely available to the public for use without a copyright notice.  
restrictions cannot be placed on its present or future use.

Although all reasonable efforts have been taken to ensure the accuracy and
reliability of the software, the United States Department of Agriculture Agricultural 
Research Service (USDA-ARS) and the U.S. Government does not and cannot warrant
the performance or results that may be obtained by using this software.  
The USDA-ARS and the U.S. Government disclaims all warranties as to performance,
merchantability or fitness for any particular purpose.

In any work or product derived from this material, proper attribution of the
authors as the source of the software should be made.



=head1 NAME

snpfindV14.0.8.pl

=head1 SYNOPSIS	

	perl snpfindV14.0.8.pl 
	
=head1 DESCRIPTION

SNPfindV14.0.8.pl expects to find the input files "../InputFiles/BlastOut" and 
"../InputFiles/BlastDB" in the indicated subdirectories. The file BlastOut is the
output file obtained by running BLASTN with the -m4 output option. The file BlastDB is
the file containing all the FASTA formated sequences used in the BLASTN analysis. 
Output files from this program are written to the ../OutputFiles/SNPfind subdirectory.
Also note that a considerable amount of commentary is written to the terminal. You
may want to capture the commentary for quality control purposes.

This program is designed to take FASTA formatted EST sequences and identify potential 
gene sequences represented in the sequences. This is accomplished in the following
steps:

=over 3

=item 1

EST libraries of a single highly inbred genome are assembled in to a single FASTA file.
All sequences are identified using their Genbank accession number.

=item 2

These ESTS are assembled into contigs using Cap3 with -o21 -f10 settings.

=item 3

The contig sequences are added to the FASTA file produced above.

=item 4

If EST sequences are available from other geneotypes, they are coded with a leading "XX" added
to their accession numbers ie AB123456 would be XXAB123456.  These sequences are then added to 
the FASTA file created in the third step.
 
=item 5

This FASTA file is then formatted with formatdb for use in BLASTN analysis.  

=item 6

The BLAST database produced above is then queried with the Cap3 consensus sequences produced in
the second step using BLASTN.  The options used are -e 1e-9 -m 4 -F F.  We want to align all of the sequence,
even the low complexity parts.  We will filter the output later to remove extraneous hits.  The
output has to be in the m4 format for efficient parsing.

=item 7

The BLAST output file and the FASTA formatted version of the BLAST database is then used as input
into the snpfind program.

=back

=head1 ALGORITHM GENERAL DESCRIPTION

This program uses a redundancy criterion for identifying sequence differences.  Thus we must see
a particular sequence difference at least twice to belive the difference and include it as a 
gene defining sequence difference (LDP).  All LDPs in a sequence make up its "haplotype" or 
gene defining pattern (LDP pattern). The program processes the alignment portion of each 
BLAST report.  Each EST hit has its aligned length calculated and is then compared to its total
EST length.  Any hit which is represented by less than 90% of its total length is removed from
consideration.  The remaining hits are processed for LDPs.  Any EST hit which has a unique LDP
pattern is validated against any coded sequences (XX) if available.  If the sequence is not 
validated, it will be removed from consideration as being a possible sequencing error.  All
remaining EST hits will be printed to the output file.

=cut 

########################################################
# Code Begins
########################################################



use strict;

my $infile = "../InputFiles/BlastOut";#BLAST output file (must be in m4 format);
my $dbfile = "../InputFiles/BlastDB";#Fasta file used to make the BLAST database
my $out = "../OutputFiles/SNPfind/snp_align_out.txt";#Parsed blast alignment
my $out2 = "../OutputFiles/SNPfind/snp_rejects.txt";#EST sequences removed for cause
my $out3 = "../OutputFiles/SNPfind/snp_W_align";#alignment of only inbred sequences
my $out4 = "../OutputFiles/SNPfind/snp_out.txt";#list of all Polymorphic characters
my $out5 = "../OutputFiles/SNPfind/snp_singletons.txt";#list of all validated singleton sequences
my $out6 = "../OutputFiles/SNPfind/snp_est_seqs.txt";#list of all validated sequences for a contig group
my $out7 = "../OutputFiles/SNPfind/snp_repetitive";#list of sequences which may have a repetitive feature


open DB, "<$dbfile" or die "Can not open Blast Database file $dbfile: $!\n";
open INFILE, "<$infile" or die "Can not open Blast Output file $infile: $!\n";
open OUT1, ">$out" or die "Can not open $out snp_align_out.txt : $!\n";
open OUT2, ">$out2" or die "Can not open $out2 : $!\n";
open OUT3, ">$out3" or die "Can not open $out3 : $!\n";
open OUT5, ">$out4" or die "Can not open $out4 : $!\n";
open OUT6, ">$out5" or die "Can not open $out5 : $!\n";
open OUT8, ">$out6" or die "Can not open $out6 : $!\n";
open OUT9, ">$out7" or die "Can not open $out7 : $!\n";



my (%lines, @lines, $line, %seqhash, @keys, $name, $seq1, $len, $record, $Qname, $char);
my ($section, $seqin, $key, $query, $ct, %snps, %seen, $hashseq, $queryseq, $querychar);
my (@snpkeys, $npercent, $snpkey, $querysize, $setback, $n, %ns,$singletons,$seq,$total);
my ($estlist,$dashct,$charct,%blanks,$counter,%ends,$value,%singleSNP,%singleton,%snpout);
my ($pos,@bases,$nt,%estlist,%base,%baselist,$accn,%db,%rep,%disc,$singletonCt);
my ($indels,$index,$gap,$ct2,%gaplist,%nwhash,$keyct);

######################################################################################
# Load the entire BLAST EST database into %db with accn number and sequence length
######################################################################################

#Go through the BLAST database and collect all the EST names and their length
#We do this so we can calculate the percent of the EST in a HIT later
#Change record separator to ">"
$/= ">";

#throw away the first ">"
<DB>;

#Go through the library fasta file record by record
while ($record = <DB>) {
	chomp($record);
	
#Use sub Fasta2Seq to split the record into name and sequence
	($name,$seq) = &Fasta2Seq($record);
	
	#try to find the accession number in the defline
	#if found, load the accession number and sequence into hash %db
	if ($name =~/gi\|\d+\|..\|.+\|([A-Z][A-Z]\d+)\s[A-Za-z]/){
		$accn = $1;
		

		#load hash %db with accession number and sequence length if the accession number is found
		$db{$accn} = length($seq);

		} #end if $name =~
	#parses Arabidopsis ests
	elsif ($name =~/gi\|\d+\|[a-z][a-z]\|[A-Z][A-Z]\d+\.\d\|([A-Z][A-Z]\d+)\s/){
			$accn = $1;
			$db{$accn} = length($seq);
		}
	elsif ($name =~ /gi\|\d+\|[a-z][a-z]\|.+\|(\S+)\s/){
			$accn = $1;
			$db{$accn} = length($seq);
		}
		#Do nothing with Contig sequences
	elsif ($name =~ /(Contig)/){
			$accn = $1;
			}
	elsif ($name =~ /^(\S+)\s+/){
		$accn = $1;
		$db{$accn} = length($seq);
		}
	else {
		#if the accession number or identifier is not found in the defline something is wrong
		#print a message to stdout to inform user
		print "Missed the identifier of $name\n";
		die "Program stopped execution because it missed an identifier for a FastA sequence in the BLAST database file\n";
		}	
		

	
		
			

}#end while
close DB;
print "Library loaded\n";

=head2 BLAST Parsing

	Function: 	Parses the M4 BLAST output
				If BLAST is not adequate for your needs, this section can be re-written/substituted
				for a parsing routine to parse your favorite alignment.  The alignment is saved
				in the global hash %seqhash.  This hash is wiped out at the end of processing
				each BLAST report.
					
=cut

###################################################################################################
# Go through the BLAST report and parse the M4 alignment from each one and process for SNPs
###################################################################################################

print "Processing Blast Report\n";
####### CHANGE THE RECORD SEPARATOR FOR PARSING YOUR BLAST/ALIGNMENT REPORTS AS NEEDED ################
$/ = "BLASTN 2.2.8 [Jan-05-2004]"; # THIS IS THE RECORD SEPARATOR FOR THE BLAST/ALIGNMENT REPORTS

#throw away the first record (its empty)
<INFILE>;


$query = "QUERY";

while ($record = <INFILE>) {
	chomp($record);
	%seqhash = ();
	$ct = 0;

	$querysize = 0;
	@keys = ();
	%snps = ();
	%seen = ();
	%singleSNP = ();
	%singleton = ();
	%blanks = ();
	$key = "";
	%nwhash = ();
	@lines = ();
	%estlist = ();
	$indels = 0;
	$singletonCt = 0;
	%rep = ();
	%disc = ();
	%gaplist = ();
	

	#look for the query sequence name and assign it to $Qname
	if ($record =~ /Query=\s(.+)/) {
		$Qname = $1;
		print "Query name is:$Qname\n";
		} #end if
		
	#search the whole record for the lines beginning with QUERY and ending with 2 spaces and Database
	if ($record =~ /\n(QUERY?\s.*\s\d+)\n\s\sDatabase/s) {
		$section = $1; #if found assign to $section
		} #end if
	else {
		print "##### Nothing found in record $Qname\n";
		next;
		} #end else

	if ($record =~ /No hits/){
		print "$Qname has no hits\n";
		}
		
	#split the alignment section ($section) into @lines
	@lines = split(/\n/, $section);
	

	#Go through each line of @lines
	foreach $line(@lines) {
		$ct = 0;

		#if the line contains nothing or if it contains "Contig"
		# then get the next line			
		if ($line eq "" || $line =~ /Contig\d+/) {
			next;
			} #end if
	
		#Extract the name field by taking the contents of the first
		#word from each line - may have to change to fit your particular situation
		if ($line =~/^(\S+)\s+/){
			$name = $1;
			} else {
				print "Missed the name of $line\n";
				}
			
		#Determine where the sequence alignment starts on each line
		#
		if ($name =~/^QUERY/) {
			$line =~/(^.{8}\s+\d+\s+)(\w|-)/;
			$setback = $1;
			$len = length($setback);
			
			} #end if
				
	#Extract the sequence field by taking  
	#the $len + 1 character to the character flanked by a space
	#and one or more digits followed by the new line character.
	#This is necessary because the alignment changes due to
	#the increase in digits of the position part of line 

		if ($line =~/^.{$len}(.+)\s\d+$/){
				$seqin = $1;
			} 
		elsif ($line =~/^.{$len}(.{60})\s\d+$/){
				$seqin = $1;
			}
		elsif ($line =~/^.{$len}(\w+\s+)\s\d+$/){
				$seqin = $1;
			}
		elsif ($line =~/^.{$len}(\s+-+)$/){
				$seqin = $1;
			}				
		else {print "##### $name missed $line\n";
			}
			
	#If the line begins with QUERY add $segin to seqhash{name}
	#add 1 to hash ($seen) to count the number of times QUERY has
	#been seen		
		if ($name =~ /^QUERY/) {
			#print "query1 is $seen{$query}\n";
			$seen{$name} += 1;
			#print "query2 is $seen{$query}\n";
			$seqhash{$name} .= $seqin;
			next;
			} #end if $name

	#Test to see if the sequence name has been seen before
		elsif ($seqhash{$name}) {
				$seen{$name} += 1;
				
	#IF the name has been seen before determine if it has been seen as many
	#times as the QUERY
				#print "Seen query = $seen{$query}  Seen $name = $seen{$name}\n";
				if ($seen{$query} > $seen{$name}) {
					my $len = length($seqhash{$name});
					
				#if QUERY has been seen more than $name will have to
				#put spaces in the middle of the sequence to keep in register
					for (my $i = $seen{$name}; $i < $seen{$query}; $i++) {
					#add 60 spaces to the sequence for n - 1 times
					#that QUERY has been seen
						$seqhash{$name} .= '                                                            '; #60 spaces
						#count how many times this happens
						$ct += 1;
						} #end for $i
					
				#Add the number of times blanks added to the sequence
				#to the times $name was seen
					$seen{$name} += $ct;
					$disc{$name}++;
					#next;
					} #end if $seen{$query} > $seen{$name}
				
				#if $name has been seen more times than the Query it must
				#have a repetative feature so skip the repetitive feature	
				if ($seen{$name} > $seen{$query}) {
					
					$rep{$name}++;
					#Need to subtract 1 from the count to restore the count
					#necessary to parse all of the sequence
					$seen{$name} -= 1;
					next;
					} #end if
				
				#Add current seqin to $seqhash{$name}
				$seqhash{$name} .= $seqin;
						
			} #end elsif
		else {
	#this is the first time the sequence name has appeared
	#need to determine how many spaces need to go on the front of the sequence
	#to keep the alignment in register
				 											
	#If the query count is greater than 1
			if ($seen{$query} > 1) {
			
		#Add 60 spaces for each time query has been seen
		#to pad the front of the sequence with the appropriate number
		#of blank spaces to keep alignment
				for (my $i = $seen{$query}; $i > 1; $i--) {
									
		$seqhash{$name} .= '                                                            '; #60 spaces
				} #end for $i
				
				$seqhash{$name} .= $seqin;
			#bring the number of times $name seen up to the number of
			#times QUERY seen
				$seen{$name} = $seen{$query};
				
				} 
			else {
				#This is the first time $name has been seen in the
				#first group in the alignment so just add seqin
				#to $seqhash{$name}
					$seqhash{$name} .= $seqin;
					#add 1 to the number of times $name was seen
					$seen{$name} += 1;
			} #end if $seen{$query}

		} #end if $name =~/QUERY/
		
	} #end foreach $line

	
	print "############### hash made for $Qname #################\n";
	
	@keys = keys(%seqhash);
	
	my @repet = keys(%rep);
	my $repetnum = $#repet + 1;
	print "Number of repetitive sequences in $Qname = $repetnum\n";
	$repetnum = $#keys;
	print "Number of sequences in $Qname = $repetnum\n";	
		
	foreach $name(@keys){
		#if the name is in %rep (repetitive sequences)
		#Add it to the list of sequences with repetitive features (OUT9)
		if ($rep{$name}){
			print OUT9 "$Qname\t$name\t$seqhash{$name}\n";
			}#end if
	
	#############################################################################
	# Inspect each subject sequence for internal spaces and change them to dashes
	#############################################################################
		#get the sequence of $name
		$seq = "";
		$seq = $seqhash{$name};
		if(!$seq){
			next;
			}
		#Go through entire sequence looking for internal spaces
		while ($seq=~/[a-z-](\s+)[a-z-]/sg){
	
			#get the spaces
			my $gapstring = $1;
			#Determin how many there are
			my $gaplen = length($gapstring);
			#Determine where they start in the sequence
			my $gapstart = pos($seq) - ($gaplen + 1);
			#Replace the spaces in $gapstring with dashes
			#print "$name Gapstring=$gapstring\n";
			$gapstring =~ s/\s/-/g;
			#print "$name Gapstring=$gapstring\n";
			#Replace the spaces with dashes
			substr($seq,$gapstart,$gaplen) = $gapstring;
			#print "$name has a deletion\n";
			#Replace old value of $seq with new $seq in %seqhash
			$seqhash{$name}=$seq;
		
		}#end while
	
	}#end foreach $name
	
	#get the new list of names from %seqhash	
	@keys = keys(%seqhash);
	
	my $hashnum = $#keys;#note do not want to count QUERY
	print "Number of names before 90% = $hashnum";
	
	##############################################################################
	#Go through each EST name and sequence and determine if 90% of the EST
	#sequence length was in the Hit
	##############################################################################
	
=head2 Alignment Processing

	Function:	Since we BLASTED using the -F F switch, any alignment could 
			contain many short "motif" hits from sequences which are
			not related to the sequence family under consideration.  So
			we must remove them from consideration if we are to identify
			meaningful LDPs. Additionally, sequences with internal repetitive 
			motifs are problematic. The program parses out the alignment of the
			first (usually the longest) alignment of a sequence with a repetitive
			element and ignores any additional alternate alignments. This routine
			may be specific to BLAST but if an alternative alignment program is used
			it should not interfere with the functioning of the rest of the program
				
=cut

	
	foreach $key(@keys){
		#skip the QUERY sequence
		if ($key eq $query){
			next;
			}
		else {
			
			$seq = $seqhash{$key};
			#remove all spaces and "-"
			$seq =~ s/[\s-]//g;
			my $hitlen = length($seq);
			if ($db{$key}){
				} else {
					print "not in FastA database $key\n";
					die "Program stopped because it incountered a sequence\nin the blast report that was not in the FastA formatted database ($key)\n";
					}
			my $minlen = $db{$key}*.90;
			
			#if $hitlen is less than $minlen then delete the record
			#from %seqhash
			if ($hitlen < $minlen){
				print OUT2 "$Qname\t$key\tlength\n";
				
				delete $seqhash{$key};
			}#end if $hitlen
			
		}#end if $key == $query
		
	}#end foreach $key
	
	#print OUT1 "############## $Qname ###################\n";
	
	#Get the new version of the name list (@keys)
	@keys = keys(%seqhash);
	
	$hashnum = $#keys;
	print " AFTER 90% = $hashnum\n";

	
	#Check %seqhash for 3 or more sequences, if it has only two sequences, we
	#cannot process the alignment any further
	# So skip the alignment and get the next record
	if (keys %seqhash < 3){
		print "$Qname has one sequence - skipping\n";
		print OUT1 "less than 2 sequences\n";
		print OUT2 "$Qname\t\tless than 2 sequences\n";
		next;
	}
			
	
	
#########################################################################################################	
#pass through the array a second time looking for positions with polymorphic characters                 #
#########################################################################################################

=head2 First Pass Analysis of the Alignment

	Function:	To initally identify sequence variations among genes within a family.
			In each alignment, the query sequence (the family consensus) should be the 
			longest of all sequences (BLAST would truncate anything past the query sequence).
			So we use its length as a maximum lenght.  We proceed column (position) by column (position)
			through the alignment and identify variants seen more than once including gap (-) 
			characters.  When we identify a sequence with a "unique LDP pattern" we attempt to validate
			it against any other geneotypes which may be included in the alignment.  If it fails
			validation, we remove it from further consideration.  We then go through the aligment and
			remove any column with only gap characters or spaces.  We consider them to be unsupported
			gaps.  Presumably these were created by sequences which were removed from consideration
			in the steps above and are now unsupported so we remove them. 
		
=cut


	#Get Query sequence and determine how long it is
	#every sequence has to be the same length
	
	$len = length($seqhash{$query});
	#print "Query is $seqhash{$query} and Len is $len\n";
	@keys = ();
	%blanks = ();
	$counter = 0;

	#get the keys to $seqhash
	@keys = keys(%seqhash);
	
	#Need to make hash %estlist with EST name and modified
	#sequence to use later.  Will need to strip all leading
	#and lagging spaces, leaving only the sequence
	
	$ct = 0;
	foreach $key(@keys) {
	

	
	#Get padded EST sequence from %seqhash and
	#assign to local variable my $seq
		my $seq = $seqhash{$key};
		
	#####load hash with all "other genotype" sequences (XX)  #############
	
		if ($key =~ /XX/){
			$nwhash{$key} = 1;
			$ct++
			}

	########################################################################
		
	#Remove all padding spaces by substituting space with nothing
		$seq =~ s/\s//g;
		
	#Load each EST sequence into holding hash %estlist ignoring the Query sequence
	#and any other genotype(XX) sequence
		unless ($key =~ /QUERY/ || $key =~ /XX/){
			$estlist{$key}=$seq;
			}#end unless
			
	}#end foreach
	if ($ct == 0){
		print "NO Other Genotype Sequences\n";
		}
	else {
		print "Number of Other Genotype Sequences = $ct\n";
		}
	$ct = 0;

	#initialize the holding hashes
	%ends = (); #hash for holding the beginning and ending position of each sequence

	my %flags = ();	#hash for holding flag values for each sequence 
					#used to determine where the sequence ends
		
	#Go through each position ($i) in each sequence and see if
	#there are more than 2 a,g,c,t,- in the column
	for (my $i = 0; $i < $len; $i++){
		my $nct = 0; #count of the n's in the sequence
		my %base = (); #Hash to keep track of EST name and the base at position $i
		my %baselist = ();#Hash to keep track of the Bases seen at position $i and the counts of each base
		my $value = "";
		my $totalbases = 0;
		my $flag = 0;

		
	#get the character from position $i in each sequence
		foreach $key(@keys){
		
		#Do not consider the QUERY sequence or Other Genotype Sequences (XX) sequences
			if ($key =~ /QUERY/ || $key =~ /XX/ ){
				next;
				}#end if
					
			if ($seqhash{$key} eq "") {
				print "### No sequence for $Qname\t$key\n";
				}#end if
			
			my $len2 = $len - 1;
		#if a character exists at $i in the sequence $key
			if ($char = substr($seqhash{$key}, $i, 1)){
				#print "Char $i in $key is $char\n";
				
				#check to see if it is the first character in the sequence
				#if it is assign $ends{$key} the value of $i
				if ($char =~/[agctn-]/ && !$ends{$key}){
					$ends{$key} = $i+1;
					#print "$Qname\t$key\tStart\t$ends{$key}\n";
					}
				elsif ($char =~ /[" """]/ && $ends{$key} && !$flags{$key}){
					#add the end position of the sequence to %ends
					my $pos = $i ;#dont need to add one to the $i value because we have gone past the end
					my $end = $ends{$key};
					$ends{$key}= "$end:$pos";
					print "$Qname\t$key\tStop\t$ends{$key}\n";
					$flags{$key} = 1; #flags that the end of the sequence has been reached
									#so we do not change the end position
					}
				#else if the est sequence ends at the end of the contig sequence
				elsif ($i == $len2 && !$flags{$key}){
					my $pos = $i + 1;
					my $end = $ends{$key};
					
					$ends{$key}= "$end:$pos";
					print "$Qname\t$key\tEND\t$ends{$key}\n";
					$flags{$key} = 1;
					}#end if $char =~/[agctn-]/

				#if the char is an a,t,g,c,n, ,-
				if ($char =~/[atgcn -]/){
					
			#set value of baselist{$key} to $char
			#keeps track of what base was associated with
			#each sequence at position $i
					$baselist{$key} = $char;
						
			#add 1 to the base counter $base{$char}
					$base{$char}++;
					}#end if$char =~ /[atgcn -]/
					
			#Else the character is unknown and we want to 
			#change it to a 'n'
				else {
					substr($seqhash{$key}, $i, 1) = 'n';
					$base{"n"}++;
					$baselist{$key} = 'n';
					}#end else $char =~
				
				}#end if $char =substr
			else {
		#the position does not exist in the sequence 
		#and we will add a space to the end of the sequence
		#which is the only place that a position could not
		#exist 
				$seqhash{$key} .= ' ';
					
			#need to add a space to the baselist{$key}
				$baselist{$key} = ' ';
						
			#need to add a space to the base count ($base{' '})
				$base{" "}++;
				
				
			#if the seq ends then must add position to %ends
				if (!$flags{$key}){
					my $pos = $i;
					my $end = $ends{$key};
					$ends{$key}= "$end:$pos";
					print "$Qname\t$key\tPosition DNE\t$ends{$key}\n";
					$flags{$key}=1;
					}#end if !$flags{$key}
				}#end else
					
		}#end foreach $key

			
		#if two or more types of bases seen and
		#one of them is not a Space
		#one type has 2 or more, add to
		#%snps
		
		#get a list of the bases seen at the position
		@bases = keys(%base);
		
		$ct = 0;
		
		#check to see if there is more than a space or an "n" and a character
		#add 1 to $ct if $nt is not a " " or a "n" or a "-"
		foreach $nt(@bases){
		
			#if ($nt eq ' ' || $nt eq "n" || $nt eq "-"){
			if ($nt eq ' ' || $nt eq "n"){
				next;
				} 
			else {
				$ct++;						
				
				}#end if $nt eq ' ' ||
					
			
				
		}#end foreach $nt(@bases)
		
		#skip the comparison if there are only one
		#type of character
		if ($ct > 1 ) {
		
		#make a list of the bases seen at the position by getting
		#the keys from %base 
			@bases = keys(%base);
			
			$nt = "";
			
		#go through each base seen
			foreach $nt(@bases){

			#if the base is a space or n or nothing get the next base
				if ($nt eq ' ' || $nt eq "n" || $nt eq ''){
					next;
					}#end if
					
		#check the number of times that base has been seen ($base{$nt})
		#if it has been seen more than 1 times set the flag
		#else find out the name(s) of the sequence(s) that had a single
		#difference 
	#########################################################################################################
	########## This step determines how many times a variant must be seen to be considered real #############
	########## Change as necessary to increase the stringency of your search                    #############
	#########################################################################################################
				if ($base{$nt} > 1) {
					next;
					}#end if
				else {
					
			#else the base was seen only 1 time
			#so go through $baselist looking for Est name with this base
					foreach $key(@keys){
			#if it is the query sequence or Other Genotype Sequence - skip it
						if ($key eq "QUERY" || $key =~ /XX/){
							next;
							}
						#add the name of the sequence with the unique base to %singleton
						#with the value of %nwhash
						if ($baselist{$key} eq $nt ) {
							if ($singleSNP{$key}){
								$singleSNP{$key} .= ",$i";
								#print "$key has $nt as a singleton at $i\n";
								}
							else {
								$singleSNP{$key}=$i;
								}
							if ($singleton{$key}){
								next;
								}
							else {
								foreach my $name(keys %nwhash){
									$singleton{$key}{$name} = 1;
									}#end foreach
								}#end if $singleton{$key}{$name}
							}
						else {
							next;
							}#end if $baselist{$key}
						}#end foreach	
					}#end if $base{$nt} >1
				}#end foreach $nt
			}#end if $ct > 1
		}#end for $i	
	################################################################################################################											
	#  Validate unique inbred line sequences 
	#############################################################################################
	
				if (%nwhash && !%singleSNP){
					print "No Singletons to validate\n";
					}
				elsif (!%nwhash && %singleSNP){
					print "No Other Genotype sequences for singletons\n";
					print "Deleting all singletons\n";
					@keys = keys (%seqhash);
					my $CtNW = @keys;
					print "Number of sequences before removing singletons = $CtNW\n";

					@keys = keys (%singleSNP);
					foreach $key(@keys){
						delete $seqhash{$key};
						print OUT2 "$Qname\t$key\tUnvalidated\n";
						}
					@keys = keys (%seqhash);
					$CtNW = @keys;
					print "Number of sequences after removing singletons = $CtNW\n";

					}						
				elsif (%nwhash && %singleSNP){
					@keys = keys (%seqhash);
					my $CtNW = @keys;
					print "Number of sequences before validation=$CtNW\n";
					&SingletonValidator;
					@keys = keys (%seqhash);
					$CtNW = @keys;
					print "Number of sequences after validation=$CtNW\n";
					}
												
	#########################################################################################
    # Search Query Sequence for single gap (-) insertions held open by nothing - remove them#
    #########################################################################################

	#get the EST contig sequence from %lines.
	#With the query anchored alignments the query sequence is the first sequence in
	#the alignment and always named "QUERY"
	
	#get new list of names
	@keys = keys %seqhash;
	
	foreach $key(@keys){
		if ($key =~ /QUERY/){
			$query = $key;
		}#end if
	}#end foreach

	$queryseq = $seqhash{$query};
	#print "Query=$queryseq\n";

	$len = length($queryseq);
	print "Query length before single gap $len\n";
	
	my $singlect = 0;
	
	#Check the Query sequence for single gaps
	
  	my @positions = ();
	my @lengths = ();
  #Search the entire Query sequence for gaps 
  	while ($queryseq =~ /[agctn]\-[agctn]/g){
  		
  #Determine the start position of the gap
  		$pos = pos($queryseq) - 2;
  		#print "$pos\t";
  #put the values into arrays
    	push(@positions, $pos);
    	
  	}#end while $seq =~
#print "\n";
if (@positions){
	#Reverse the order of the arrays so we start with the right-most 
	#gap first - necessary so the position information is not thrown off
  		@positions = reverse @positions;
	#Get the number of indexes from @positions
  		my $index = @positions;
  		print "Single Gap Index=$index\n";
	#Go through each index
  		for (my $i = 0;$i<$index;$i++){
    		$pos = $positions[$i];
		
			#print "Postion is $pos\n";
			#select one character at position $pos
			my $charct = 0;
			my $total = 0;
			my $dashct = 0;
	
		
			#go through each sequence at that position and count the gaps "-"
			foreach $key(@keys) {

			#Extract the sequence for the key and assign it to $hashseq
				$hashseq = $seqhash{$key};
				
				unless(length($hashseq) >= $pos){
					next;
					}
					
			#assign character $pos from each sequence to $char if it exists
				if ($char = substr($hashseq, $pos, 1)){
					#print "dash char is : $char at position $pos\n";
					$total++;
					} #end if $char
				else {
			#If the character does not exist, get the next key
					next;
				} #end else
			
				#print "Char$pos=$char\n";
							
			#If the character is a "-" add 1 to the dash count ($dashct)
				if ($char eq "-"){
					$dashct++;
				} #end if $char
					
			#if $char is a [a,g,c,t,n] add 1 to character count ($charct)
				if ($char =~ /[agctn]/) {
					$charct++;
				} #end if $char
					
			} #end foreach $key


			#if the character count ($charct) ==0 remove the gap and the character 
			#causing the gap - since only one, we are assuming it is a sequencing
			#error 
					
			if ($charct == 0) {
				$hashseq = $seqhash{$query};
				$len = length($hashseq);
				
				foreach $key(@keys) {

					$hashseq = $seqhash{$key};
					unless(length($hashseq) >= $pos){
						next;
						}
					
				#If character $pos exsists in $hashseq, replace it with nothing (removing it)
					if ($char = substr($hashseq, $pos, 1) ){
						#print "Deleting $key:$char\n";
						substr($hashseq, $pos, 1) = '';
					} #end if substr($hashseq						
					$seqhash{$key} = $hashseq;
					
				} #end foreach $key
			
				$hashseq = $seqhash{$query};
				$len = length($hashseq);
	
			
			} #end if $charct == 0
		
		}#end for my $i
		
	}#end if @positions
	
	######################################################################################
	$hashseq = $seqhash{$query};
	#print "Query=$hashseq\n";
	
	#clear out old %snps hash because we have a new alignment to look through
	%snps = ();
	#clear out old %ends hash because the 3prime end will change
	%ends = ();
	%flags = ();
	$len = length($hashseq);
	my $len2 = $len - 1;
	#wipe out old %estlist
	%estlist = ();
	$keyct = 0;
	
	####### Pass through the remaining inbred sequences looking for LDP's ##############

=head2 Final Identification of LDPs

	Function:	To identify the LDPs remaining after removing the extraneous gap characters and
			removing the unvalidated sequences.  Therefore, we go through the alignment as
			above and identify all the gene defining sequence differences (LDPs) left
			in the alignment.  Note that this is a redundant list which will be used in the
			program SortHaplotypes.pl to make a non-redundant list of all potential haplotypes
			(genes) in each gene family.

=cut


	foreach $name(@keys){
		
		$seq = $seqhash{$name};
		#Print the alignments to snp_align_out.txt
		#print OUT1 "$Qname\t$name\t$seq\n";

		
		#remove leading and lagging spaces
		$seq =~ s/\s//g;
		#Load each EST sequence into holding hash %estlist ignoring the Query sequence
		#and Other Genotype Sequences (XX)
		unless ($name =~ /QUERY/ || $name =~ /XX/){
			$estlist{$name}=$seq;
			}#end unless

	
	# Remove all Other Genotype Sequence ests else print sequence info to inbred sequence
	# alignment (snp_W_align)
		if ($name =~ /XX/){
			#delete $seqhash{$name};
			next;
			}
		else {
			$keyct++;
			print OUT3 "$Qname\t$name\t$seqhash{$name}\n";
			}
		}#end foreach
		
	### check to see if there are any inbred sequences left don't count query
	@keys = keys (%seqhash);
	#my $keyct = @keys;
	unless ($keyct > 1) {
		print "No Inbred sequences remaining after validation\n";
		print OUT8 "$Qname\t \t \t \n";
		print OUT6 "$Qname\t \t \n";
		next; #get next blast record
		}
		
	#Go through each position ($i) in each sequence and see if
	#there are more than 2 a,g,c,t,- in the column
	for (my $i = 0; $i < $len; $i++){
		my $nct = 0; #count of the n's in the sequence
		my %base = (); #Hash to keep track of EST name and the base at position $i
		my %baselist = ();#Hash to keep track of the Bases seen at position $i and the counts of each base
		my $value = "";
		my $totalbases = 0;
		my $flag = 0;

	#get the character from position $i in each sequence
		foreach $key(@keys){
		
		#Do not consider the QUERY sequence or Other Genotype Sequence
			if ($key =~ /QUERY/ || $key =~ /XX/){
				next;
				}
					
			if ($seqhash{$key} eq "") {
				print "### No sequence for $Qname\t$key\n";
				}
			
		#if a character exists at $i in the sequence $key
			if ($char = substr($seqhash{$key}, $i, 1)){
				
	
	########################################################################
	
					#check to see if it is the first character in the sequence
				#if it is assign $ends{$key} the value of $i
				if ($char =~/[agctn-]/ && !$ends{$key}){
					$ends{$key} = $i+1;
					
					}
				elsif ($char =~ /[" """]/ && $ends{$key} && !$flags{$key}){
					#add the end position of the sequence to %ends
					my $pos = $i ;#dont need to add one to the $i value because we have gone past the end
					my $end = $ends{$key};
					$ends{$key}= "$end:$pos";
					print "$Qname\t$key\tStop\t$ends{$key}\n";
					$flags{$key} = 1; #flags that the end of the sequence has been reached
									#so we do not change the end position
					}
				#else if the est sequence ends at the end of the contig sequence
				elsif ($i == $len2 && !$flags{$key}){
					my $pos = $i + 1;
					my $end = $ends{$key};
					
					$ends{$key}= "$end:$pos";
					print "$Qname\t$key\tEND\t$ends{$key}\n";
					$flags{$key} = 1;
					}#end if $char =~/[agctn-]/

	
	
	#########################################################################
									
			#set value of baselist{$key} to $char
			#keeps track of what base was associated with
			#each sequence at position $i
					$baselist{$key} = $char;
						
			#add 1 to the base counter $base{$char}
					$base{$char}++;
				
								
				}#end if (substr


			}#end foreach

			
		#if two or more types of bases seen and
		#one of them is not a Space
		#one type has 2 or more, add to
		#%snps
		
		#get a list of the bases seen at the position
		@bases = keys(%base);
		
		$ct = 0;
		
		#check to see if there is more than a space or an "n" and a character
		#add 1 to $ct if $nt is not a " " or a "n" or a "-"
		foreach $nt(@bases){
		
			if ($nt eq ' ' || $nt eq "n" || $nt eq ""){
				next;
				}
				########################################################
				# test $nt for "-" if found save gap start position
				########################################################
			if ($nt eq "-" && $#bases > 0){
				$ct++;
				#go through all the names looking for the ones with a "_"
				foreach $key(@keys){
					if ($key eq "QUERY"||$key =~ /XX/){
						next;
						}
					if ($baselist{$key} eq $nt ) {
						if ($gaplist{$key}){
							next;
							}
						else {
							$gaplist{$key}= $i+1;
								
							}#end if $gaplist
							
						}#end if $baselist
		
					}#end foreach $key
				} 
			else {
				$ct++;						
				
				}#end if $nt eq "-"
						
		}#end foreach $nt(@bases)
		
		#skip the comparison if there are only one
		#type of character
		if ($ct > 1 ) {
		
		#make a list of the bases seen at the position by getting
		#the keys from %base 
			@bases = keys(%baselist);
			
			$nt = "";
			$flag = 1;

			$n = "";

		#if the $flag has been set then the position has a snp and
		#needs to be added to %snps	
			#print "### Flag is:$flag \n";
			if ($flag == 1){
				foreach $key(@keys){
					#Skip the query sequence or Other Genotype Sequences
					if ($key eq "QUERY"||$key =~ /XX/){
						next;
						}#end if $key
					my $char = $baselist{$key};
				#if the character is a space or n or nothing get the next base
					if ($char eq ' ' || $char eq ''){
						next;
						}#end if
							
				#if $key is in $gaplist, check to see if $nt is a "-"
				#if not then put in a Deletion as next snp and delete the key
				#from $gaplist
					if ($gaplist{$key} && $char ne "-"){
						my $gapstart = $gaplist{$key};
						my $gapend = $i + 1 ;
						$value = "$gapstart:D:$gapend";
						$snpout{$key} .= "$value,";
						$snps{$key} .= "$value,";
						delete $gaplist{$key};
						}#end if $gaplist
												
					my $char2 = $char;
				#if the character is something expected [agctn\s-]
					if(!$gaplist{$key} && $char =~ /[agctn\s-]/){
						my $pos = ($i + 1);
						$char2 =~ tr/[agctn]/[AGCTN]/;
						$value = "$pos:$char2:$base{$char}";
						#print "Regular SNP value is $value\n";
						#$n = "$pos:$npercent,";
						$snps{$key} .= "$value,";
						$snpout{$key} .= "$char";
						}
					elsif ($gaplist{$key} && $char eq "-"){
						next;
						}  
					else {
						#else the character is unexpected so give $value
						# a distinctive string
						$value = "$i: :Z";
						$snps{$key} .= "$value,";
						$snpout{$key} .= "$char";
						}
				}#end foreach $key
						
			}#end if $flag == 1
			
			
			}#end if ct > 1
		else {
			foreach $key(@keys){
					
				my $char = $baselist{$key};
					
			#if $key is in $gaplist, check to see if $char is a "-"
			#if not then put in a Deletion as next snp and delete the key
			#from $gaplist
				if ($gaplist{$key} && $char ne "-"){
					my $gapstart = $gaplist{$key};
					my $gapend = $i + 1 ;
					$value = "$gapstart:D:$gapend";
					$snpout{$key} .= "$value,";
					$snps{$key} .= "$value,";
					delete $gaplist{$key};
					}#end if $gaplist
				}#end foreach $key
			}#end else
		
		#Keep count of all positions	
		$counter++;

	}#end for my $i
	
#########################################################################################
# Print out sequences in %seqhash #
#########################################################################################

my $Qlen = length($seqhash{$query});

#Print Qname and Qlen to OUT3 (snp_Qlengths.txt);
print  "$Qname\t$Qlen\n";
#print OUT9 "$Qname\t$Qlen\n";


#initalize the sequence counter $ct
$ct = 0;

#initalize the singleton counter
$singletonCt = 0;

#Initalize and declare the snp counter snpCt
my $snpCt = 0;

#Declare and Initialize $flag5
my $flag5 = 0;

#Declare and Initalize temporary holder for snp_est_seq.txt 
my $templine = "";
my $templine2 = "";

#Get the names of all sequences in %seqhash	
@keys = keys(%seqhash);

#Get the names of all sequences with SNPs
@snpkeys = keys(%snps);
	
	
#####################################################
# Change any sequence with SNPs to UC at snp position
#####################################################

=head2 Sequence Annotation for LDPs

	Function:	To annotate each sequence to easily visualize the positions with
			LDPs (gene defining sequence differences).  To accomplish this we
			go through each sequence and capitalize the character in each LDP position.
			This makes it easy to identify the sequence differences in an aligment by hand.
			Upper case letters are used in the subsequent programs as an indicator of a LDP.
			So this must not be changed or the programs will break.

=cut 


if (%snps){
	foreach $snpkey(@snpkeys){
		#skip the Query sequence and any singletons
		if ($snpkey =~/QUERY/ || $snpkey =~ /XX/) {
			next;
			}
		my @snps2 = ();
		my $snpline = $snps{$snpkey};
		@snps2 = split(/,/,$snpline);
		my $end = $ends{$snpkey};
		my ($pos5,$pos3) = split(/:/,$end);

		foreach  my $snp(@snps2){
			($pos,$char,my $num) = split(/:/,$snp);
			
			#if the sequence ends before the first snp stop here
			if ($pos3 < $pos) {last;}
			
			#if the snp occurs before the begining of the sequence get the
			#next snp
			if ($pos < $pos5) {next;}
			
			#if $char is a space, skip it else process the snp
			if ($char =~/[AGCTN-]/){
				my $newpos = $pos - ($pos5-0);
				$len = length($estlist{$snpkey});
				
				if ($pos < $len + $pos5){
					my $char2 = substr($estlist{$snpkey},$newpos,1);
					$char2 =~ tr/[agctn]/[AGCTN]/;									
					if ($char2 ne $char){
						print "##### $snpkey $char2 DNE $char at $newpos\n";
						substr($estlist{$snpkey},$newpos,1) = $char2;
						substr($seqhash{$snpkey},$pos,1) = $char2;
						} 
					else {
						#print "$snpkey  pos is $newpos  Char $char2:$char\n";
						substr($estlist{$snpkey},$newpos,1) = $char;
						substr($seqhash{$snpkey},$pos,1) = $char;
						}#end else of if $char2
					} 
				else {
					print "###Bad $snpkey\tlen is $len\t pos is $newpos\tEnd is $pos3\n";
					}#end else of if $pos <
				
				} #end if $char =~
			
			}#end foreach my $snp
		
		}#end foreach $snpkey
	
	}#end if %snps


#####################################################	

	#Go through each name in %seqhash 
	foreach $name(@keys){
	
		my $end = "";
		my $pos5 = "";
		my $pos3 = "";
	
	#Print the alignments to snp_align_out.txt
		print OUT1 "$Qname\t$name\t$seqhash{$name}\n";

	#if the name is QUERY skip it
		if ($name =~/QUERY/ || $name =~ /XX/){
			next;
			}

	#Get the start and stop position for $name
		$end = $ends{$name};
		($pos5,$pos3) = split(/:/,$end);

		$len = $pos3-$pos5;
		#print "Name=$name, Pos5=$pos5, Pos3=$pos3\n";

		#if the name had snps associated with it and it is not a singleton
		if ($snps{$name}){
			print OUT5 "$Qname:$name\t$snps{$name}$len\t$pos5\t$pos3\n";
			$templine .= "$name,$estlist{$name},$pos5,$pos3\t";
			next;
			}
		else {
		#$name did not have any snps so we write it out unchanged to
		#snp_est_seqs.txt(8) and snp_out.txt(5)
		#need to check to see if $name exists in %snps if it does
		#then we print it out with $snps{$name} else we print it with a blank

				print OUT5 "$Qname:$name\t\t$pos5\t$pos3\n";
				$templine2 .= "$name,$estlist{$name},$pos5,$pos3\t";
				next;
			}#end if $snps
		
		}#end foreach $name
	if ($templine eq "" && $templine2 ne ""){
		
		print "print to OUT8-1\n";
		print OUT8 "$Qname\t$templine2$Qlen\n";
		}
		
	if ($templine ne "" && $templine2 eq ""){
		print OUT8 "$Qname\t$templine$Qlen\n";
		}
		
	if ($templine ne "" && $templine2 ne ""){
		print OUT8 "$Qname\t$templine$templine2$Qlen\n";
		}



	
}#end of $record while statement

my ($year,$month,$day) = &GetDate;
		
print "SNP Find Version 14.0.8 is Done !!!!!\n$month-$day-$year\n";
close OUT1;
close OUT2;
close OUT9;
close OUT5;	
close OUT6;

close OUT8;
close OUT3;
exit;		

#Sub Fasta2Seq
#
#Subroutine to convert a Fasta record into
#its definition line and sequence
#
#Input is a Fasta record
#
#Output is Definition line, Sequence
#

sub Fasta2Seq {

my $record = "";
my $DefLine = "";
my $seq = "";
my $line = "";
my @lines = {};

$record = shift(@_);

@lines = split(/\n/, $record);
	
$DefLine = shift(@lines);
	
$seq = join("", @lines);

return $DefLine,$seq;

}

################################################################
# Subroutine GetDate - Composes a date stamp
################################################################

sub GetDate {
my $year = 0;
my $month = 0;
my $day = 0;
#get date information for a day stamp
($day,$month,$year)=(localtime)[3,4,5];
return $year+1900, $month+1, $day;

}#end sub GetDate

	
#
#Subroutine to validate est's with unique polymorphisms
#
#
#

=head2 Unique Sequence Validation

	Function:	To attempt to validate a sequence which is unique to the inbred
			genotype sequences but was represented in other genotype sequences.
			This is an attempt to recover gene sequences of genes which were expressed
			poorly in the tissues/conditions of the cDNA libraries obtained from the
			inbred genotype by expanding the analysis to other genotypes if available.
			Each unique sequence difference observed in a sequence is compared to the 
			character in the same position in any other geneotype sequence contained in 
			the alignment.  If a matching character is found in any non-inbred sequence
			that character is considered validated.  The process is continued until a
			non-inbred sequence is found which validates ALL the unique characters in 
			the sequence under consideration. If none is found, then the sequence is
			considered un-validated and removed from further consideration.

=cut


sub SingletonValidator() {
	
	#get list of all singletons from %singleSNP
	my @keys = keys(%singleSNP);
	my $snpnum = @keys;
	print "Number of singletons before validation=$snpnum\n";
	my @nwnames = keys(%nwhash);
	my $nwct = @nwnames;
	print "Number of NonWill names = $nwct\n";
	
	#foreach singleton name
	foreach $key(@keys){
		print "Processing singleton $key\n";
		my %nwhash2 = ();
		my @nwhash2 = ();
		foreach my $name(@nwnames){
			$nwhash2{$name}= 1;
			}
		@nwhash2 = keys(%nwhash2);
		#print "nwhash2 names= @nwhash2\n";
		#get a list of all singleton positions by splitting up
		#$singleSNP line
		my @SNPpos = split /,/, $singleSNP{$key};
		#Go through each SNP position in the singleton
		foreach my $pos(@SNPpos){
			my $char = substr($seqhash{$key},$pos,1);
			my $posFlag = 0;
			$nwct = 0;
			#print "Char=$char\t";
			#and compare the value at $pos for each inbred line sequence
			#get all inbred line names

			@nwnames = keys(%nwhash2);
			foreach my $nwname (@nwnames) {
				#print "$nwname\t$pos\t";
				#unless ($seqhash{$nwname}){
					#next;
					#}
				#print "$nwname\t$seqhash{$nwname}\n";
				my $nwlen = length($seqhash{$nwname});
				if ($nwlen < $pos){
					next; #skip the name
					print "$key\t$nwname skipped\n";
					}
				if (my $char2 = substr($seqhash{$nwname},$pos+1,1)){

					if ($char =~ /[nN ]/ || $char2 =~ /[nN ]/){
				 		next; #skip if either char is a space or "n"
				 		}
					#print "$nwname\t$pos\tChar=$char\tChar2=$char2\n";
					#if the characters match go to the next Other Genotype Sequence name
					if ($char eq $char2) {
						$posFlag = 1;
						next;
						}
					else {
						#delete the Other Genotype Sequence name from the list
						delete $singleton{$key}{$nwname};
						delete $nwhash2{$nwname};
						$nwct--;
						#print "Deleted $nwname\n";
						}#end if $char

						
					}#end if my $char2
				
				}#end foreach $nwname
				
				#print "posFlag=$posFlag\n";
			unless ($posFlag == 1){
				print OUT2 "$Qname\t$key\tUnvalidated\n";
				#print OUT6 "$Qname\t$key\n";
				delete $seqhash{$key};
				delete $singleSNP{$key};
				print "Deleted singleton $key - No match to a unique character\n";
				last;
				}

			}#end foreach $pos
		$nwct = keys(%nwhash2);
		#print "nwct=$nwct\n";
		if ($nwct == 0){
				print OUT2 "$Qname\t$key\tUnvalidated\n";
				#print OUT6 "$Qname\t$key\n";
				delete $seqhash{$key};
				delete $singleSNP{$key};
				print "-------- could not validate a character\n";
				
				#last;
				}
			
		}#end foreach $key
	@keys = keys(%singleSNP);
	$snpnum = @keys;
	print "Number of singletons after validation=$snpnum\n";
	if ($snpnum > 0){
		foreach $name(@keys){
			print OUT6 "$Qname\t$name\n";
			}
		}
		
}#end sub SingletonValidator

